# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Utilities for TC filenaming, for use within geoips filename formatters."""

import logging
from datetime import datetime, timedelta
from glob import glob
from os.path import basename as pathbasename
from os.path import join as pathjoin
from geoips.interfaces import products

LOG = logging.getLogger(__name__)

# interface = None indicates to the GeoIPS interfaces that this is not a valid
# plugin, and this module will not be added to the GeoIPS plugin registry.
# This allows including python modules within the geoips/plugins directory
# that provide helper or utility functions to the geoips plugins, but are
# not full GeoIPS plugins on their own.
interface = None


def get_storm_subdir(
    basin_path, base_tc_stormname, tc_stormnum, output_dict, sector_info
):
    """Get the TC storm subdirectory."""
    # Default to just the "base" tc storm name (ie, WP932022 or SH162022)
    tc_stormname = base_tc_stormname

    # If it exists, get the storm_start_datetime from the sector_info dictionary
    storm_start_datetime = None
    if sector_info and "original_storm_start_datetime" in sector_info:
        storm_start_datetime = sector_info["original_storm_start_datetime"]
    elif sector_info and "storm_start_datetime" in sector_info:
        storm_start_datetime = sector_info["storm_start_datetime"]

    # Now check if any "file_path_modifications" were specified in the output config.
    # This will determine if we need a more specific/unique subdirectory for a specific
    # storm.
    # Ie, WP932022.2022020506 vs WP932022
    if (
        isinstance(output_dict, dict)
        and "file_path_modifications" in output_dict
        and isinstance(storm_start_datetime, datetime)
        and tc_stormnum >= 69
    ):
        path_mods = output_dict["file_path_modifications"]
        if "unique_invest_dirs" in path_mods and path_mods["unique_invest_dirs"]:
            tc_stormname = storm_start_datetime.strftime(
                f"{base_tc_stormname}.%Y%m%d%H"
            )
            LOG.info("SETTING unique storm dir %s", tc_stormname)
        if (
            "existing_invest_dirs_allowable_time_diff" in path_mods
            and path_mods["existing_invest_dirs_allowable_time_diff"]
        ):
            base_path = pathjoin(basin_path, base_tc_stormname)
            paths = glob(base_path + "*")
            # Make the minimum greater than the allowable - so we can track the lowest
            min_timediff = abs(
                timedelta(days=path_mods["existing_invest_dirs_allowable_time_diff"])
            )
            for path in paths:
                path_parts = pathbasename(path).split(".")
                if len(path_parts) > 1:
                    try:
                        existing_storm_start_datetime = datetime.strptime(
                            path_parts[1], "%Y%m%d%H"
                        )
                    except ValueError:
                        LOG.warning(
                            "SKIPPING using existing invest dir, "
                            "no valid start datetime in path %s",
                            path,
                        )
                        continue
                    timediff = abs(storm_start_datetime - existing_storm_start_datetime)
                    if timediff < min_timediff:
                        LOG.info(
                            "SETTING unique storm dir to existing %s, %s < %s",
                            path_parts[1],
                            timediff,
                            min_timediff,
                        )
                        min_timediff = timediff
                        tc_stormname = pathbasename(path)
                    else:
                        LOG.info(
                            "SKIPPING not using existing dir %s, %s > %s",
                            path_parts[1],
                            timediff,
                            min_timediff,
                        )

        LOG.info("USING final storm dir %s", tc_stormname)
    return tc_stormname


def tc_storm_basedir(
    basedir, tc_year, tc_basin, tc_stormnum, output_dict=None, sector_info=None
):
    """Produce base storm directory for TC web output.

    Parameters
    ----------
    basedir : str
         base directory
    tc_year : int
         Full 4 digit storm year
    tc_basin : str
         2 character basin designation
            SH Southern Hemisphere
            WP West Pacific
            EP East Pacific
            CP Central Pacific
            IO Indian Ocean
            AL Atlantic
    tc_stormnum : int
        2 digit storm number
            90 through 99 for invests
            01 through 69 for named storms

    Returns
    -------
    path : str
        Path to base storm web directory
    """
    tc_date = "tc{0:04d}".format(tc_year)
    base_tc_stormname = "{0}{1:02d}{2:04d}".format(tc_basin, tc_stormnum, tc_year)
    basin_path = pathjoin(basedir, tc_date, tc_basin)
    tc_storm_subdir = get_storm_subdir(
        basin_path, base_tc_stormname, tc_stormnum, output_dict, sector_info
    )
    # Need to actually add a check to see if an invest directory exists that is "close"
    # If one does, use that instead of a new one (which will result in
    # creating the new directory downstream)
    path = pathjoin(basin_path, tc_storm_subdir)
    return path


def update_extra_field(
    output_dict,
    xarray_obj,
    area_def,
    product_name,
    extra_field_delimiter="_",
    existing_extra_field=None,
    extra_field_provider=True,
    extra_field_coverage_func=False,
    extra_field_resolution=False,
    include_filename_extra_fields=False,
):
    """Finalize extra field using standard geoips arguments."""
    if extra_field_provider is True:
        from geoips.filenames.base_paths import PATHS as gpaths

        extra_field_provider = gpaths["GEOIPS_COPYRIGHT_ABBREVIATED"]

    from geoips.dev.product import get_covg_args_from_product

    prod_spec_override = None
    if output_dict is not None:
        prod_spec_override = output_dict.get("product_spec_override")

    if "source_names" in xarray_obj.attrs:
        for source_name in xarray_obj.source_names:
            try:
                prod_plugin = products.get_plugin(
                    source_name, product_name, prod_spec_override
                )
                covg_args = get_covg_args_from_product(
                    prod_plugin,
                    covg_field="filename_coverage_checker",
                )
            except KeyError:
                continue
    else:
        prod_plugin = products.get_plugin(
            xarray_obj.source_name,
            product_name,
            prod_spec_override,
        )
        covg_args = get_covg_args_from_product(
            prod_plugin,
            covg_field="filename_coverage_checker",
        )

    extras = []

    if extra_field_provider:
        extras += [f"{extra_field_provider}"]
    if extra_field_coverage_func and "radius_km" in covg_args:
        extras += [f"cr{covg_args['radius_km']}"]
    if (
        include_filename_extra_fields
        and "filename_extra_fields" in xarray_obj.attrs
        and xarray_obj.filename_extra_fields
    ):
        for field in xarray_obj.filename_extra_fields:
            extras += [f"{xarray_obj.filename_extra_fields[field]}"]

    # This must go first - but no "res" if no other fields
    if extra_field_resolution:
        resolution = max(area_def.pixel_size_x, area_def.pixel_size_y) / 1000.0
        res_str = "{0:0.1f}".format(resolution).replace(".", "p")
        if len(extras) > 0:
            res_str = f"res{res_str}"
        extras = [res_str] + extras

    # This doesn't count towards "other fields" for the res in resolution.
    if existing_extra_field:
        extras += [f"{existing_extra_field}"]

    extra = extra_field_delimiter.join(extras)
    return extra



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `test sector` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

from os import environ

import pytest

from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


class TestGeoipsTestSector(BaseCliTest):
    """Unit Testing Class for Test Sector Command."""

    @property
    def command_combinations(self):
        """A list of call signatures for the GeoipsTestSector command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            base_args = ["geoips", "test", "sector"]
            self._cmd_list = []
            # Only creating two sectors here as they'll hit the same code locations.
            # Just want to make sure we get 100% coverage and don't create a ton of new
            # files.
            self._cmd_list.append(base_args + ["goes_east"])
            self._cmd_list.append(
                base_args + ["japan", "--outdir", f"{environ['GEOIPS_OUTDIRS']}"]
            )
            # Add argument list to retrieve help message
            self._cmd_list.append(base_args + ["-h"])
            # Add argument list with non existent sector
            self._cmd_list.append(base_args + ["non_existent_sector"])
        return self._cmd_list

    def check_error(self, args, error):
        """Ensure that the 'geoips test sector ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        assert "To use, type `geoips test sector <sector_name>" in error

    def check_output(self, args, output):
        """Ensure that the 'geoips test sector ...' successful output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        # The args provided are valid, so test that the output is actually correct
        if "-h" in args:
            assert "To use, type `geoips test sector <sector_name>" in output
        else:
            assert f"{args[3]}.png" in output


test_sub_cmd = TestGeoipsTestSector()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips test sector ...' commands.

    This test covers valid combinations of commands for the 'geoips test sector'
    command. We also test invalid commands, to ensure that the proper help documentation
    is provided for those using the command incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with (ie. ['geoips', 'test', 'sector'])
    """
    test_sub_cmd.test_command_combinations(monkeypatch, args)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Matplotlib-based clean image output."""

import logging

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "image"
name = "imagery_clean"


def call(
    area_def,
    xarray_obj,
    product_name,
    output_fnames,
    product_name_title=None,
    mpl_colors_info=None,
    existing_image=None,
    remove_duplicate_minrange=None,
    fig=None,
    main_ax=None,
    mapobj=None,
):
    """Plot clean image on matplotlib figure."""
    success_outputs = []
    plot_data = xarray_obj[product_name].to_masked_array()
    plot_kwargs = {}
    if hasattr(xarray_obj, "fuse_order"):
        plot_kwargs = {"zorder": int(xarray_obj.fuse_order)}
    from geoips.image_utils.mpl_utils import create_figure_and_main_ax_and_mapobj
    from geoips.image_utils.colormap_utils import set_matplotlib_colors_standard
    from geoips.image_utils.mpl_utils import plot_image, save_image

    if not mpl_colors_info:
        # Create the matplotlib color info dict - the fields in this dictionary
        # (cmap, norm, features, etc) will be used in plot_image to ensure the image
        # matches the colorbar.
        mpl_colors_info = set_matplotlib_colors_standard(
            data_range=[plot_data.min(), plot_data.max()],
            cmap_name=None,
            cbar_label=None,
        )

    if fig is None and main_ax is None and mapobj is None:
        # Create matplotlib figure and main axis, where the main image will be plotted
        fig, main_ax, mapobj = create_figure_and_main_ax_and_mapobj(
            area_def.x_size, area_def.y_size, area_def, noborder=True
        )

    # Plot the actual data on a map
    plot_image(
        main_ax, plot_data, mapobj, mpl_colors_info=mpl_colors_info, **plot_kwargs
    )

    if output_fnames is not None:
        for clean_fname in output_fnames:
            LOG.info("Saving the clean image %s", clean_fname)
            # Save the clean image with no gridlines or coastlines
            success_outputs += save_image(
                fig,
                clean_fname,
                is_final=False,
                image_datetime=xarray_obj.start_datetime,
                remove_duplicate_minrange=remove_duplicate_minrange,
            )

    return success_outputs



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read derived surface winds from KNMI scatterometer netcdf data."""

# Python Standard Libraries
from copy import deepcopy
from glob import glob
import logging
from os.path import basename

# Third-Party Libraries
import numpy
import xarray

# GeoIPS imports
from geoips.xarray_utils.time import (
    get_min_from_xarray_time,
    get_max_from_xarray_time,
)

LOG = logging.getLogger(__name__)

MS_TO_KTS = 1.94384
DEG_TO_KM = 111.321

interface = "readers"
family = "standard"
name = "scat_noaa_winds_netcdf"
source_names = ["ascat"]


def read_noaa_data(wind_xarray):
    """Reformat ascat xarray object appropriately.

    * variables: latitude, longitude, time,
      wind_speed_kts, wind_dir_deg_met
    * attributes: source_name, platform_name, data_provider,
      interpolation_radius_of_influence
    """
    geoips_metadata = {}
    geoips_metadata["data_provider"] = "noaa"
    # Setting standard geoips attributes
    LOG.info("Reading %s data", wind_xarray.source)

    geoips_metadata["source_name"] = "ascat"
    geoips_metadata["platform_name"] = wind_xarray.platform.lower()

    # Manually construct the time array, since xarray cannot auto decode these data
    from astropy.time import Time
    from dateutil.relativedelta import relativedelta

    utc_time = Time(wind_xarray.time_seconds, format="unix_tai").utc.datetime
    # Seconds since time is w.r.t. 1990-01-01, need to offset 20 years
    utc_time += relativedelta(years=20)

    wind_xarray["time"].data = utc_time
    wind_xarray["time"].attrs["units"] = "Time in UTC"
    wind_xarray["time"].attrs.pop("valid_min")
    wind_xarray["time"].attrs.pop("valid_max")

    # In NOAA files, pixel size is only mentioned in the title
    # e.g. L2OVW50kmASCAT
    psize_string = wind_xarray.title
    for string in ["L2OVW", "kmASCAT"]:
        psize_string = psize_string.replace(string, "")
    pixel_size = float(psize_string)

    # Interpolation Radius of Influence
    geoips_metadata["interpolation_radius_of_influence"] = pixel_size * 1000.0

    geoips_metadata["sample_distance_km"] = pixel_size

    # setting wind_speed_kts appropriately
    wind_xarray["wind_speed_kts"] = wind_xarray["wind_speed"] * MS_TO_KTS
    wind_xarray["wind_speed_kts"].attrs = wind_xarray["wind_speed"].attrs
    wind_xarray["wind_speed_kts"].attrs["units"] = "kts"

    # Save wind direction to standardized name across ascat readers
    wind_xarray["wind_dir_deg_met"] = wind_xarray["wind_dir"]
    wind_xarray["wind_dir_deg_met"].attrs = wind_xarray["wind_dir"].attrs
    wind_xarray.wind_dir_deg_met.attrs["standard_name"] = "wind_from_direction"
    wind_xarray.wind_dir_deg_met.attrs["valid_max"] = 360

    # Set lat/lons/timestamp appropriately
    # Also rename wind ambiguity keys
    wind_xarray = wind_xarray.rename(
        {
            "lat": "latitude",
            "lon": "longitude",
            "wind_speed_sol": "wind_speed_ambiguity_kts",
            "wind_dir_sol": "wind_dir_deg_ambiguity_met",
        }
    )

    RAIN_FLAG_BIT = 9
    if hasattr(xarray, "ufuncs"):
        # ufuncs no longer available as of xarray v2022.06 (at least)
        # ufuncs appears to be back as of v2025.3.1
        wind_xarray["rain_flag"] = xarray.ufuncs.logical_and(
            wind_xarray["wvc_quality_flag"], (1 << RAIN_FLAG_BIT)
        )
        # rf variable used downstream for determining ambiguities
        rf = wind_xarray["rain_flag"]
    else:
        # Can not figure out how to do the logical and in xarray now -
        # so do it in numpy.
        # This may not be the most efficient, especially if we are trying to use
        # dask / lazy processing.
        # Dropping the ".to_masked_array()" appears to lose the nan values -
        # but could perhaps do that then re-mask?
        rf = numpy.logical_and(
            wind_xarray["wvc_quality_flag"].to_masked_array(), (1 << RAIN_FLAG_BIT)
        )
        data_dims = {}
        for dim in wind_xarray["wvc_quality_flag"].dims:
            data_dims[dim] = wind_xarray.dims[dim]
        wind_xarray["rain_flag"] = xarray.DataArray(
            rf.astype(int), coords=wind_xarray.coords, dims=data_dims
        )

    # Create rain flag for ambiguity products
    # The rain_flag variable is only 2D. We need to
    # create a rain_flag_ambiguity variable that has
    # the same dimensions as "wind_speed_ambiguity_kts"
    ambs = wind_xarray["wind_speed_ambiguity_kts"]
    # Create data_dims dictionary for ambiguity variable
    data_dims = {x: wind_xarray.dims[x] for x in ambs.dims}
    # Create 3D rain flag by stacking 2D rain_flag by number of ambiguities
    rf_amb = numpy.dstack([rf] * data_dims["MaxAmbiguity"])
    # Set 3D rain flag values to NaN where ambiguities are NaN
    nan_ambs = numpy.isnan(ambs.data)
    rf_amb[nan_ambs] = numpy.nan
    # Store to xarray dataset
    wind_xarray["rain_flag_ambiguity"] = xarray.DataArray(
        rf_amb.astype(int), coords=wind_xarray.coords, dims=data_dims
    )

    wind_xarray = wind_xarray.set_coords(["time"])
    return wind_xarray, geoips_metadata


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read KNMI scatterometer derived winds from netcdf data.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    final_wind_xarrays = {}
    ingested = []
    for fname in fnames:
        wind_xarray = xarray.open_dataset(str(fname), decode_times=False)

        LOG.info("Read data from %s", fname)

        if hasattr(wind_xarray, "title") and "ASCAT" in wind_xarray.title:
            wind_xarray, geoips_metadata = read_noaa_data(wind_xarray)

        geoips_metadata["source_file_names"] = [basename(fname)]
        geoips_metadata["minimum_coverage"] = 20
        ingested += [(wind_xarray, geoips_metadata)]

    for curr_xobj, curr_geoips_metadata in ingested:
        curr_file_metadata = deepcopy(curr_xobj.attrs)
        curr_xobj.attrs = {**curr_file_metadata, **curr_geoips_metadata}

        if "WINDSPEED" not in final_wind_xarrays:
            final_xobj = curr_xobj.copy()
            final_xobj.attrs = curr_geoips_metadata
            final_xobj.attrs["source_file_attributes"] = [curr_file_metadata]
            final_xobj.attrs["source_file_names"] = curr_geoips_metadata[
                "source_file_names"
            ].copy()
            final_wind_xarrays["WINDSPEED"] = final_xobj
            continue

        final_xobj = final_wind_xarrays["WINDSPEED"]

        for attr_name in ["source_name", "platform_name"]:
            if final_xobj.attrs[attr_name] != curr_xobj.attrs[attr_name]:
                raise ValueError(
                    f"Attribute '{attr_name}' must match on all data files"
                    f"Expected '{final_xobj.attrs[attr_name]}',"
                    f"Current value '{curr_xobj.attrs[attr_name]}'"
                )
        final_xobj = xarray.concat([final_xobj, curr_xobj], dim="NUMROWS")
        final_xobj.attrs["source_file_names"] += curr_geoips_metadata[
            "source_file_names"
        ]
        final_xobj.attrs["source_file_attributes"] += [curr_file_metadata]
        final_wind_xarrays["WINDSPEED"] = final_xobj

    for wind_xarray in final_wind_xarrays.values():
        LOG.info("Setting standard metadata")
        wind_xarray.attrs["start_datetime"] = get_min_from_xarray_time(
            wind_xarray, "time"
        )
        wind_xarray.attrs["end_datetime"] = get_max_from_xarray_time(
            wind_xarray, "time"
        )

        if "wind_speed_kts" in wind_xarray.variables:
            # These text files store wind speeds natively in kts
            wind_xarray["wind_speed_kts"].attrs["units"] = "kts"

        LOG.info(
            "Read data %s start_dt %s source %s platform %s data_provider %s roi "
            "%s native resolution",
            wind_xarray.attrs["start_datetime"],
            wind_xarray.attrs["source_name"],
            wind_xarray.attrs["platform_name"],
            wind_xarray.attrs["data_provider"],
            wind_xarray.attrs["interpolation_radius_of_influence"],
            wind_xarray.attrs["sample_distance_km"],
        )

    final_wind_xarrays["METADATA"] = wind_xarray[[]]

    return final_wind_xarrays


def get_test_files(test_data_dir):
    """Generate test xarray from test files for unit testing."""
    filepath = (
        test_data_dir
        + "/test_data_scat/data/20230524_metopc_"
        + "noaa_class_tc2023wp02mawar/L2OVW25kmASCAT_v1r1_m03_s202305242348000*.nc"
    )
    filelist = glob(filepath)[:2]

    tmp_xr = call(filelist)
    if len(filelist) == 0:
        raise NameError("No files found")
    return tmp_xr


def get_test_parameters():
    """Generate test data key for unit testing."""
    return [{"data_key": "WINDSPEED", "data_var": "wind_speed_kts"}]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Sector adjusters interface module."""

from geoips.interfaces.base import BaseModuleInterface


class SectorAdjustersInterface(BaseModuleInterface):
    """Interface for adjusting a sector based on additional information.

    Ie, pass in data and the area definition, and modify the area_definition based
    on the data itself.
    """

    name = "sector_adjusters"
    required_args = {
        "list_xarray_list_variables_to_area_def_out_fnames": [
            "xobjs",
            "area_def",
            "variables",
        ]
    }
    required_kwargs = {
        "list_xarray_list_variables_to_area_def_out_fnames": ["recenter_variables"]
    }


sector_adjusters = SectorAdjustersInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Passive Microwave 37 GHz Polarization Corrected Temperature.

Data manipulation steps for the "37pct" product.
This algorithm expects Brightness Temperatures in units of degrees Kelvin.
"""
import logging

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "list_numpy_to_numpy"
name = "pmw_37pct"


def call(
    arrays,
    output_data_range=None,
    min_outbounds="crop",
    max_outbounds="mask",
    norm=False,
    inverse=False,
):
    """37pct product algorithm data manipulation steps.

    This algorithm expects Brightness Temperatures in units of degrees Kelvin,
    and returns degrees Kelvin

    Parameters
    ----------
    arrays : list of numpy.ndarray
        * numpy.ndarray or numpy.MaskedArray of channel data, in order of sensor
            "channels" list
        * Degrees Kelvin

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of appropriately scaled channel data,
        in degrees Kelvin.
    """
    h37 = arrays[0]
    v37 = arrays[1]

    out = (2.15 * v37) - (1.15 * h37)

    if output_data_range is None:
        output_data_range = [230.0, 280.0]

    from geoips.data_manipulations.corrections import apply_data_range

    data = apply_data_range(
        out,
        min_val=output_data_range[0],
        max_val=output_data_range[1],
        min_outbounds=min_outbounds,
        max_outbounds=max_outbounds,
        norm=norm,
        inverse=inverse,
    )
    return data



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Test retrieving GeoIPS and other plugin packages' plugins."""
from importlib import reload
import pytest

from geoips import interfaces
from geoips.errors import PluginError
from geoips import plugin_registry as plugin_registry_module


def yield_interface_plugin_tuples():
    """Yield tuples containing information for one call to `test_get_plugin`.

    Yield tuples (interface_name, plugin_name_tuple, expected_output) to use when
    attempting to load a plugin in the `test_get_plugin` function. The tuple will
    contain three elements:

    - interface_name:
        - The name of the interface that we want to attempt to load a plugin from.
    - plugin_name_tuple:
        - A tuple containing one or two strings defining which plugin should be
            loaded. Two strings is used for the Products interface and one string is
            used for all other interfaces.
    - expected_output:
        - A string describing the expected outcome. This is used in the
            `test_get_plugin` function to determine which part of the logic to follow
            and how to test the results.
    """
    registry = plugin_registry_module.plugin_registry.registered_plugins
    for plugin_type in registry:
        if plugin_type == "text_based":
            # Functionality is not yet set up for this. Once PR #504 gets merged, we can
            # remove this.
            continue
        for interface_name in registry[plugin_type]:
            if interface_name == "products":
                # Products have another level which is 'source_name', so we need
                # different functionality here to handle that
                for source_name in registry[plugin_type][interface_name]:
                    products = registry[plugin_type][interface_name][source_name]
                    for idx, prod_name in enumerate(products):
                        # Only add up to the first 5 plugins of the first source name
                        if idx > 4:
                            break
                        yield (interface_name, (source_name, prod_name), "normal")
                    break
            else:
                for idx, plugin_name in enumerate(
                    registry[plugin_type][interface_name]
                ):
                    # Only add up to the first 5 plugins of this interface
                    if idx > 4:
                        break
                    yield (interface_name, (plugin_name,), "normal")
    # Test that retrieving a fake plugin raises a PluginError
    yield ("algorithms", ("fake_algorithm_plugin",), "no_plugin")
    reload(plugin_registry_module)
    # Test that retrieving a fake plugin without rebuilding the registry raises a
    # PluginError
    yield ("algorithms", ("another_fake_algorithm_plugin",), "no_rebuild")
    registry = plugin_registry_module.plugin_registry.registered_plugins
    single_channel_entry = registry["module_based"]["algorithms"].pop("single_channel")
    registry["module_based"]["algorithms"][
        "out_of_sync_single_channel"
    ] = single_channel_entry
    # Test that an out of sync plugin causes 'create_plugin_registries' to be ran so it
    # is synced up again
    yield ("algorithms", ("single_channel",), "out_of_sync")
    reload(plugin_registry_module)
    registry = plugin_registry_module.plugin_registry.registered_plugins
    registry["module_based"]["algorithms"]["single_channel"][
        "relpath"
    ] = "some/fake/path"
    # Test that pointing a fake path to an existing plugin resets the registry and the
    # path is fixed
    yield ("algorithms", ("single_channel",), "non_existent_path")


def gen_label(val):
    """Generate the interface-plugin_name-<test_result> id for pytest."""
    return f"{val[0]}-{val[1]}-{val[2]}"


@pytest.mark.parametrize(
    "int_plug_tuple", yield_interface_plugin_tuples(), ids=gen_label
)
def test_get_plugin(int_plug_tuple):
    """Test the functionality of 'get_plugin' for the provided interface and plugin."""
    interface_name = int_plug_tuple[0]
    plugin_name = int_plug_tuple[1]
    test_result = int_plug_tuple[2]
    curr_interface = getattr(interfaces, interface_name)
    # All of these should result in a plugin getting returned. This assumes we've been
    # given a valid plugin name
    if test_result in ["normal", "out_of_sync", "non_existent_path"]:
        assert curr_interface.get_plugin(*plugin_name, rebuild_registries=True)
    # If this is in fact a plugin that does not exist, a plugin error will be raised.
    elif test_result == "no_plugin":
        with pytest.raises(PluginError):
            curr_interface.get_plugin(*plugin_name, rebuild_registries=True)
    elif test_result == "no_rebuild":
        with pytest.raises(PluginError):
            curr_interface.get_plugin(*plugin_name, rebuild_registries=False)



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Test Order-based procflow ReaderArgumentsModel."""

import copy

# Third-Party Libraries
import pytest
from pydantic import ValidationError

# GeoIPS Libraries
from geoips.pydantic import workflows


def test_good_valid_reader_arguments_model(valid_reader_arguments_model_data):
    """Tests ReaderArgumentsModel with valid inputs."""
    model = workflows.ReaderArgumentsModel(**valid_reader_arguments_model_data)

    assert model.area_def == "None"
    assert model.variables == ["None"]
    assert model.metadata_only is True
    assert model.self_register == ["None"]
    assert model.fnames == ["None"]


def test_bad_reader_arguments_model_field_defaults():
    """Tests if the default value for a given field is valid or not."""
    model = workflows.ReaderArgumentsModel()

    expected_defaults = {
        "area_def": None,
        "variables": None,
        "metadata_only": False,
        "self_register": None,
        "fnames": None,
    }

    for field, expected_value in expected_defaults.items():
        assert (
            getattr(model, field) == expected_value
        ), f"Default value for '{field}' is incorrect"


def test_bad_reader_arguments_model_invalid_field_type():
    """Tests ReaderArgumentsModel for invalid field type instantiation."""
    invalid_test_data = {
        "area_def": 123,
        "chans": "channel_1",
        "metadata_only": "invalid",
        "self_register": "invalid",
        "fnames": "invalid",
    }

    with pytest.raises(ValidationError) as exec_info:
        workflows.ReaderArgumentsModel(**invalid_test_data)

    error_info = exec_info.value.errors()

    test_data_errors = {
        "area_def": "string_type",
        "chans": "list_type",
        "metadata_only": "bool_parsing",
    }
    for field, error_type in test_data_errors.items():
        assert any(
            err["loc"] == (field,) and err["type"] == error_type for err in error_info
        ), f"Expected error for the field '{field}' with the type '{error_type}'."


def test_bad_reader_arguments_model_empty_list_for_variable(
    valid_reader_arguments_model_data,
):
    """Tests ReaderArgumentsModel with empty list input for variable field."""
    valid_reader_arguments_model_data["variables"] = []

    model = workflows.ReaderArgumentsModel(**valid_reader_arguments_model_data)

    assert model.variables == []


def test_bad_reader_arguments_model_chans_deprecation(
    caplog, valid_reader_arguments_model_data
):
    """Tests ReaderArgumentsModel with valid inputs."""
    invalid_test_data = copy.deepcopy(valid_reader_arguments_model_data)
    invalid_test_data["chans"] = invalid_test_data.pop("variables")

    with caplog.at_level("WARNING"):
        model = workflows.ReaderArgumentsModel(**invalid_test_data)

    # Assert model still created
    assert model

    # Check that a warning was logged
    assert any("'chans' is deprecated" in msg for msg in caplog.messages)



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Workflow interface module."""

import logging

from geoips.interfaces.base import BaseYamlInterface
from geoips.pydantic.workflows import WorkflowPluginModel


LOG = logging.getLogger(__name__)


class WorkflowsInterface(BaseYamlInterface):
    """Interface for workflow plugins."""

    name = "workflows"
    validator = WorkflowPluginModel


workflows = WorkflowsInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Module containing tpw_pwat ASCII palette based colormap."""

import logging

LOG = logging.getLogger(__name__)

interface = "colormappers"
family = "matplotlib"
name = "tpw_pwat"


def call():
    """Colormap for displaying data using TPW PWAT ascii colormap.

    Data range of ASCII palette is 1 to 90 mm, with numerous transitions

    Returns
    -------
    mpl_colors_info : dict
        Dictionary of matplotlib plotting parameters, to ensure consistent
        image output

    See Also
    --------
    :ref:`api`
        ASCII palette is found in image_utils/ascii_palettes/tpw_pwat.txt
    """
    # levels in mm (evenly spaced by 2 mm except for 1-6):
    values = [
        1,
        2,
        3,
        4,
        5,
        6,
        8,
        10,
        12,
        14,
        16,
        18,
        20,
        22,
        24,
        26,
        28,
        30,
        32,
        34,
        36,
        38,
        40,
        42,
        44,
        46,
        48,
        50,
        52,
        54,
        56,
        58,
        60,
        62,
        64,
        66,
        68,
        70,
        72,
        74,
        76,
        78,
        80,
        82,
        84,
        86,
        88,
        90,
    ]

    from geoips.image_utils.colormap_utils import from_ascii
    from matplotlib.colors import BoundaryNorm
    from geoips.geoips_utils import find_ascii_palette

    bounds = values + [values[-1] + 1]
    mpl_cmap = from_ascii(find_ascii_palette(name))

    mpl_colors_info = {
        "cmap": mpl_cmap,
        "norm": BoundaryNorm(bounds, mpl_cmap.N),
        "cbar_ticks": values,
        "cbar_tick_labels": None,
        "cbar_label": r"TPW (mm)",
        "boundaries": None,
        "cbar_spacing": "uniform",
        "colorbar": True,
        "cbar_full_width": True,
    }

    return mpl_colors_info



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read derived surface winds from SAR netcdf data."""

# Python Standard Libraries
from datetime import datetime
import glob
import logging
from os.path import basename

# Third-Party Libraries
import numpy
import xarray

# GeoIPS imports
from geoips.xarray_utils.time import (
    get_min_from_xarray_time,
    get_max_from_xarray_time,
)

LOG = logging.getLogger(__name__)

MS_TO_KTS = 1.94384
DEG_TO_KM = 111.321

interface = "readers"
family = "standard"
name = "sar_winds_netcdf"
source_names = ["sar-spd"]


def read_sar_data(wind_xarray):
    """Reformat SAR xarray object appropriately.

    * variables: latitude, longitude, time, wind_speed_kts
    * attributes: source_name, platform_name, data_provider,
      interpolation_radius_of_influence
    """
    # Setting standard geoips attributes
    LOG.info("Reading SAR data")
    wind_xarray.attrs["source_name"] = "sar-spd"
    if (
        "platform" in wind_xarray.attrs
        and wind_xarray.platform.lower() == "sentinel-1a"
    ):
        wind_xarray.attrs["platform_name"] = "sentinel-1"
    elif "platform" in wind_xarray.attrs and wind_xarray.platform.lower() == "rcm-1":
        wind_xarray.attrs["platform_name"] = "rcm-1"
    elif "platform" in wind_xarray.attrs and wind_xarray.platform.lower() == "rcm-2":
        wind_xarray.attrs["platform_name"] = "rcm-2"
    elif "platform" in wind_xarray.attrs and wind_xarray.platform.lower() == "rcm-3":
        wind_xarray.attrs["platform_name"] = "rcm-3"
    elif (
        "platform" in wind_xarray.attrs and wind_xarray.platform.lower() == "radarsat-2"
    ):
        wind_xarray.attrs["platform_name"] = "radarsat-2"
    else:
        raise ValueError(
            f"Unsupported satellite name for SAR data: {wind_xarray.platform}"
        )
    wind_xarray.attrs["interpolation_radius_of_influence"] = 3000
    # For resampling to a minimum-sized grid
    wind_xarray.attrs["sample_distance_km"] = 3.0
    wind_xarray.attrs["sample_pixels_x"] = 300
    wind_xarray.attrs["sample_pixels_y"] = 300
    wind_xarray.attrs["minimum_coverage"] = 0
    wind_xarray.attrs["granule_minutes"] = 0.42
    wind_xarray.attrs["data_provider"] = "star"
    if "acknowledgment" in wind_xarray.attrs and "NOAA" in wind_xarray.acknowledgment:
        wind_xarray.attrs["data_provider"] = "star"
    # Used for tc filenames / text files

    LOG.info("Shape: %s", wind_xarray["sar_wind"].shape)
    # setting wind_speed_kts appropriately
    wind_xarray["wind_speed_kts"] = wind_xarray["sar_wind"] * MS_TO_KTS
    wind_xarray["wind_speed_kts"].attrs = wind_xarray["sar_wind"].attrs
    wind_xarray["wind_speed_kts"].attrs["units"] = "kts"
    wind_xarray["wind_speed_kts"] = xarray.where(
        wind_xarray.mask == -1, wind_xarray.wind_speed_kts, numpy.nan
    )

    # Set lat/lons appropriately
    wind_xarray = wind_xarray.rename({"latitude": "latitude", "longitude": "longitude"})

    # Set time appropriately
    # Get the full array of times.  pandas is much better with time series.
    wind_pandas = wind_xarray.to_dataframe()

    # This worked with xarray version 0.16.1
    # wind_xarray['time'] = wind_pandas['acquisition_time'][:, 5, :].to_xarray()
    #                                                                       .transpose()

    # Make sure it grabs the right ones, no matter which order x, y, and xfit are in.
    # Later versions (0.18.0) of xarray can have different orders -
    # ensure we use labels, not explicit locations
    time_array = (
        wind_pandas["acquisition_time"]
        .to_xarray()[dict(x=slice(None), y=slice(None), xfit=5)]
        .transpose("y", "x")
    )
    # Remove the xfit coordinate - no longer needed
    wind_xarray["time"] = time_array.reset_coords("xfit", drop=True)

    wind_xarray = wind_xarray.set_coords(["time"])
    wind_xarray["sigma"] = xarray.where(
        wind_xarray.sigma == 0, numpy.nan, wind_xarray.sigma
    )
    # This is not correct - should be -35 to -20, need to find documentation
    # for deriving NRCS from sigma
    wind_xarray["nrcs"] = 10 * numpy.log10(wind_xarray.sigma)
    wind_xarray["incident_angle"] = wind_xarray["incid"]
    return [wind_xarray]


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read SAR derived winds from netcdf data.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    # Only SAR reads multiple files
    fname = fnames[0]
    wind_xarray = xarray.open_dataset(str(fname))
    wind_xarray.attrs["data_provider"] = "unknown"
    wind_xarray.attrs["source_name"] = "unknown"
    wind_xarray.attrs["platform_name"] = "unknown"
    wind_xarray.attrs["interpolation_radius_of_influence"] = "unknown"
    wind_xarray.attrs["source_file_names"] = [basename(fname)]
    wind_xarray.attrs["sample_distance_km"] = "unknown"

    LOG.info("Read data from %s", fname)

    if (
        hasattr(wind_xarray, "source")
        and ("SAR" in wind_xarray.source or "SAR" in wind_xarray.title)
        and hasattr(wind_xarray, "title")
    ):
        wind_xarrays = []
        columns = None
        for fname in fnames:
            LOG.info("    Reading file %s", fname)
            wind_xarray = xarray.open_dataset(str(fname))
            LOG.info(
                "        rows: %s, columns: %s", wind_xarray.rows, wind_xarray.columns
            )
            if columns is None:
                columns = wind_xarray.columns
            if columns == wind_xarray.columns:
                wind_xarrays += read_sar_data(wind_xarray)
            else:
                LOG.info("            COLUMNS DOES NOT MATCH, NOT APPENDING")
        if len(fnames) == 1:
            wind_xarrays = {"WINDSPEED": wind_xarrays[0]}
        else:
            final_xarray = xarray.Dataset()
            lat_array = xarray.DataArray(
                numpy.vstack(
                    [
                        curr_xarray.latitude.to_masked_array()
                        for curr_xarray in wind_xarrays
                    ]
                )
            )
            lon_array = xarray.DataArray(
                numpy.vstack(
                    [
                        curr_xarray.longitude.to_masked_array()
                        for curr_xarray in wind_xarrays
                    ]
                )
            )
            time_array = xarray.DataArray(
                numpy.vstack(
                    [curr_xarray.time.to_masked_array() for curr_xarray in wind_xarrays]
                )
            )
            wspd_array = xarray.DataArray(
                numpy.vstack(
                    [
                        curr_xarray.wind_speed_kts.to_masked_array()
                        for curr_xarray in wind_xarrays
                    ]
                )
            )
            sigma_array = xarray.DataArray(
                numpy.vstack(
                    [
                        curr_xarray.sigma.to_masked_array()
                        for curr_xarray in wind_xarrays
                    ]
                )
            )
            final_xarray["latitude"] = lat_array
            final_xarray["longitude"] = lon_array
            final_xarray["time"] = time_array
            final_xarray["wind_speed_kts"] = wspd_array
            final_xarray["sigma"] = sigma_array
            final_xarray.attrs = wind_xarrays[0].attrs

            wind_xarrays = {"WINDSPEED": final_xarray}

    for wind_xarray in wind_xarrays.values():
        if not hasattr(wind_xarray, "minimum_coverage"):
            wind_xarray.attrs["minimum_coverage"] = 20

        LOG.info("Setting standard metadata")
        wind_xarray.attrs["start_datetime"] = get_min_from_xarray_time(
            wind_xarray, "time"
        )
        wind_xarray.attrs["end_datetime"] = get_max_from_xarray_time(
            wind_xarray, "time"
        )

        if "wind_speed_kts" in wind_xarray.variables:
            # These text files store wind speeds natively in kts
            wind_xarray["wind_speed_kts"].attrs["units"] = "kts"

        LOG.info(
            "Read data %s start_dt %s source %s platform %s data_provider %s roi "
            "%s native resolution",
            wind_xarray.attrs["start_datetime"],
            wind_xarray.attrs["source_name"],
            wind_xarray.attrs["platform_name"],
            wind_xarray.attrs["data_provider"],
            wind_xarray.attrs["interpolation_radius_of_influence"],
            wind_xarray.attrs["sample_distance_km"],
        )

    wind_xarrays["METADATA"] = wind_xarray[[]]
    if wind_xarrays["METADATA"].start_datetime == wind_xarrays["METADATA"].end_datetime:
        # Use alternate attributes to set start and end datetime
        try:
            wind_xarrays["METADATA"].attrs["start_datetime"] = datetime.strptime(
                wind_xarray.time_coverage_start, "%Y%m%dT%H%M%S"
            )
            wind_xarrays["METADATA"].attrs["end_datetime"] = datetime.strptime(
                wind_xarray.time_coverage_end, "%Y%m%dT%H%M%S"
            )
        except ValueError:
            # 20221103 used YYYYMMDDTHHMNSS, on 20221105 switched to
            #                                                       YYYY-MM-DDTHH:MN:SSZ
            # Allow both
            wind_xarrays["METADATA"].attrs["start_datetime"] = datetime.strptime(
                wind_xarray.time_coverage_start, "%Y-%m-%dT%H:%M:%SZ"
            )
            wind_xarrays["METADATA"].attrs["end_datetime"] = datetime.strptime(
                wind_xarray.time_coverage_end, "%Y-%m-%dT%H:%M:%SZ"
            )

    return wind_xarrays


def get_test_files(test_data_dir):
    """Generate testing xarray from test data."""
    filepath = test_data_dir + "/test_data_sar/data/*.nc"
    filelist = glob.glob(filepath)
    tmp_xr = call(filelist)
    if len(filelist) == 0:
        raise NameError("No files found")
    return tmp_xr


def get_test_parameters():
    """Generate test data key for unit testing."""
    return [{"data_key": "WINDSPEED", "data_var": "wind_speed_kts"}]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read SAPHIR hdf files."""

# Python Standard Libraries
from datetime import datetime
import logging

# Third-Party Libraries
import h5py
import numpy as np
import xarray as xr

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "saphir_hdf5"
source_names = ["saphir"]


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read SAPHIR hdf data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    if len(fnames) > 1:
        raise ValueError(
            "Multiple files not supported with this reader. "
            "Please call with a single file."
        )
    fname = fnames[0]
    fileobj = h5py.File(str(fname), mode="r")

    lon_obj = fileobj["ScienceData"]["Longitude_Samples"]
    londata = float(lon_obj.attrs["scale_factor"]) * (lon_obj[...]) + float(
        lon_obj.attrs["add_offset"]
    )
    lat_obj = fileobj["ScienceData"]["Latitude_Samples"]
    # Offset is -40.0, scale_factor is 0.01
    latdata = float(lat_obj.attrs["scale_factor"]) * (lat_obj[...]) + float(
        lat_obj.attrs["add_offset"]
    )

    date_time_group = fileobj["ScienceData"]["Scan_FirstSampleAcqTime"][...]

    sstart_date_time_group = date_time_group[0, 0]
    # NOTE in Python 3, since h5py returns a binary string, splitting on ' ' failed,
    # but splitting on the default delimeter worked.
    # This appears to work on Python 2 as well.
    # sstart_date_group, sstart_time_group = sstart_date_time_group.split(' ')
    sstart_date_group, sstart_time_group = sstart_date_time_group.split()
    sstart_group_time = sstart_time_group[0:6]

    start_date_group = int(sstart_date_group)
    start_time_group = int(sstart_group_time)

    edtg = date_time_group.size - 1
    send_date_time_group = date_time_group[0, edtg]
    # NOTE in Python 3, since h5py returns a binary string, splitting on ' ' failed,
    # but splitting on the default delimeter worked.
    # This appears to work on Python 2 as well.
    send_date_group, send_time_group = send_date_time_group.split()
    send_group_time = send_time_group[0:6]

    end_date_group = int(send_date_group)
    end_time_group = int(send_group_time)

    incidence_angle = 0.01 * np.ma.masked_equal(
        fileobj["ScienceData"]["IncidenceAngle_Samples"], 32767
    )
    ch1qf = np.ma.masked_equal(
        np.squeeze((fileobj["ScienceData"]["QF_Samples_S1"][...])), 65535
    )
    ch2qf = np.ma.masked_equal(
        np.squeeze((fileobj["ScienceData"]["QF_Samples_S2"][...])), 65535
    )
    ch3qf = np.ma.masked_equal(
        np.squeeze((fileobj["ScienceData"]["QF_Samples_S3"][...])), 65535
    )
    ch4qf = np.ma.masked_equal(
        np.squeeze((fileobj["ScienceData"]["QF_Samples_S4"][...])), 65535
    )
    ch5qf = np.ma.masked_equal(
        np.squeeze((fileobj["ScienceData"]["QF_Samples_S5"][...])), 65535
    )
    ch6qf = np.ma.masked_equal(
        np.squeeze((fileobj["ScienceData"]["QF_Samples_S6"][...])), 65535
    )
    ch1 = np.ma.masked_equal(
        np.squeeze(0.01 * (fileobj["ScienceData"]["TB_Samples_S1"][...])), 65535
    )
    ch2 = np.ma.masked_equal(
        np.squeeze(0.01 * (fileobj["ScienceData"]["TB_Samples_S2"][...])), 65535
    )
    ch3 = np.ma.masked_equal(
        np.squeeze(0.01 * (fileobj["ScienceData"]["TB_Samples_S3"][...])), 65535
    )
    ch4 = np.ma.masked_equal(
        np.squeeze(0.01 * (fileobj["ScienceData"]["TB_Samples_S4"][...])), 65535
    )
    ch5 = np.ma.masked_equal(
        np.squeeze(0.01 * (fileobj["ScienceData"]["TB_Samples_S5"][...])), 65535
    )
    ch6 = np.ma.masked_equal(
        np.squeeze(0.01 * (fileobj["ScienceData"]["TB_Samples_S6"][...])), 65535
    )
    scanqf = np.ma.masked_equal(
        np.squeeze((fileobj["ScienceData"]["SAPHIR_QF_scan"][...])), 65535
    )

    LOG.info("Reading file %s", fname)

    # setup saphir xarray
    xarray_saphir = xr.Dataset()
    xarray_saphir["latitude"] = xr.DataArray(latdata)
    xarray_saphir["longitude"] = xr.DataArray(londata)
    xarray_saphir["IncidenceAngle"] = xr.DataArray(incidence_angle)
    xarray_saphir["scan_qf"] = xr.DataArray(scanqf)
    xarray_saphir["ch1qf_183.31_0.2"] = xr.DataArray(ch1qf)
    xarray_saphir["ch2qf_183.31_1.1"] = xr.DataArray(ch2qf)
    xarray_saphir["ch3qf_183.31_2.8"] = xr.DataArray(ch3qf)
    xarray_saphir["ch4qf_183.31_4.2"] = xr.DataArray(ch4qf)
    xarray_saphir["ch5qf_183.31_6.8"] = xr.DataArray(ch5qf)
    xarray_saphir["ch6qf_183.31_11.0"] = xr.DataArray(ch6qf)
    xarray_saphir["ch1_183.31_0.2"] = xr.DataArray(np.ma.masked_where(ch1qf > 64, ch1))
    xarray_saphir["ch2_183.31_1.1"] = xr.DataArray(np.ma.masked_where(ch2qf > 64, ch2))
    xarray_saphir["ch3_183.31_2.8"] = xr.DataArray(np.ma.masked_where(ch3qf > 64, ch3))
    xarray_saphir["ch4_183.31_4.2"] = xr.DataArray(np.ma.masked_where(ch4qf > 64, ch4))
    xarray_saphir["ch5_183.31_6.8"] = xr.DataArray(np.ma.masked_where(ch5qf > 64, ch5))
    xarray_saphir["ch6_183.31_11.0"] = xr.DataArray(np.ma.masked_where(ch6qf > 64, ch6))
    # xarray_saphir['time']=xr.DataArray(pd.DataFrame(time_scan).astype(int).apply(pd.to_datetime,format='%Y%j%H%M'))

    # add attributes to xarray
    xarray_saphir.attrs["start_datetime"] = datetime.strptime(
        "%08d%06d" % (start_date_group, start_time_group), "%Y%m%d%H%M%S"
    )
    xarray_saphir.attrs["end_datetime"] = datetime.strptime(
        "%08d%06d" % (end_date_group, end_time_group), "%Y%m%d%H%M%S"
    )
    xarray_saphir.attrs["source_name"] = "saphir"
    xarray_saphir.attrs["platform_name"] = "meghatropiques"
    xarray_saphir.attrs["data_provider"] = "GSFC"
    xarray_saphir.attrs["sample_distance_km"] = 10
    xarray_saphir.attrs["interpolation_radius_of_influence"] = 20000

    return {"SAPHIR": xarray_saphir, "METADATA": xarray_saphir[[]]}



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Interpolators interface module."""

from geoips.interfaces.base import BaseModuleInterface


class InterpolatorsInterface(BaseModuleInterface):
    """Interpolation routine to apply when reprojecting data."""

    name = "interpolators"
    required_args = {
        "2d": ["area_def", "input_xarray", "output_xarray", "varlist"],
        "grid": ["area_def", "input_xarray", "output_xarray", "varlist"],
    }

    required_kwargs = {"2d": ["array_num"], "grid": ["array_num"]}


interpolators = InterpolatorsInterface()



# # # Distribution Statement A. Approved for public release. Distribution unlimited.
# # #
# # # Author:
# # # Naval Research Laboratory, Marine Meteorology Division
# # #
# # # This program is free software: you can redistribute it and/or modify it under
# # # the terms of the NRLMMD License included with this program. This program is
# # # distributed WITHOUT ANY WARRANTY; without even the implied warranty of
# # # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the included license
# # # for more details. If you did not receive the license, for more information see:
# # # https://github.com/U-S-NRL-Marine-Meteorology-Division/

"""Routines for writing SMAP or SMOS windspeed data in AWIPS2 compatible format."""
import logging
from pathlib import Path
from datetime import datetime, timedelta
import numpy as np
import xarray as xr
from geoips.xarray_utils.time import (
    get_min_from_xarray_time,
    get_max_from_xarray_time,
    get_datetime_from_datetime64,
)
from geoips.filenames.base_paths import PATHS as geoips_variables

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "xrdict_to_outlist"
name = "windspeed_awips2_formatter"


def call(
    xarray_dict,
    working_directory=geoips_variables["GEOIPS_OUTDIRS"],
):
    """Write AWIPS2 compatible NetCDF files from SMAP or SMOS windspeed data.

    Parameters
    ----------
    xarray_dict : Dict[str, xarray.Dataset]
    working_directory : str

    Returns
    -------
    List[str]
    """
    working_dir = Path(working_directory)
    utc_date_format = "%Y-%m-%d %H:%M:%S UTC"
    success_outputs = []
    if xarray_dict["METADATA"].attrs["source_name"] == "smap-spd":
        success_outputs = write_smap_awips2(xarray_dict, working_dir, utc_date_format)
    elif xarray_dict["METADATA"].attrs["source_name"] == "smos-spd":
        success_outputs = write_smos_awips2(xarray_dict, working_dir, utc_date_format)
    else:
        LOG.warning(
            (
                "Unrecognized source_name, '%s' cannot be converted into "
                + "AWIPS2-compatible windspeed NetCDF file"
            ),
            xarray_dict["METADATA"].attrs["source_name"],
        )

    return success_outputs


def write_smap_awips2(xarray_dict, working_dir, utc_date_format):
    """Write AWIPS2 style NetCDF file containing SMAP data to the disk.

    Parameters
    ----------
    xarray_dict : dict[str, xarray.Dataset]
    working_dir : Path
    utc_date_format : str

    Returns
    -------
    List[str]
    """
    success_outputs = []
    ascending_nodes = xarray_dict["WINDSPEED_1"]
    descending_nodes = xarray_dict["WINDSPEED_2"]
    asc_times = ascending_nodes["time"]
    des_times = descending_nodes["time"]
    asc_speeds = ascending_nodes["wind_speed_kts"].values
    des_speeds = descending_nodes["wind_speed_kts"].values
    longitude = ascending_nodes["longitude"].values
    latitude = ascending_nodes["latitude"].values

    start_datetime = xarray_dict["METADATA"].attrs["start_datetime"]
    mintime, maxtime = get_min_max_times_from_two_nodes(
        ascending_nodes, descending_nodes, "time"
    )

    # create roughly one section per hour
    sections_count = int((maxtime - mintime).total_seconds() // 3600)
    seconds_per_section = int((maxtime - mintime).total_seconds() // sections_count)

    for section_index in range(sections_count):
        split_start = start_datetime + timedelta(
            seconds=seconds_per_section * section_index
        )
        split_end = split_start + timedelta(seconds=seconds_per_section)
        base_file = "CIRA_A2WIND_SMAP_" + split_start.strftime("%Y%m%d_%H%M")
        awips2_file = Path(working_dir, f"{base_file}.nc")
        if awips2_file.exists():
            LOG.warning("awips2 file already exists! Overwriting: %s", awips2_file)

        timestamp_data = np.full(asc_times.shape, -999.0)
        new_smap_data = np.full(asc_times.shape, -999.0)

        asc_mask = make_smap_data_mask(asc_times, split_start, split_end)
        new_smap_data[asc_mask] = asc_speeds[asc_mask]

        des_mask = make_smap_data_mask(des_times, split_start, split_end)
        new_smap_data[des_mask] = des_speeds[des_mask]

        for i in range(len(asc_mask[0])):
            j = asc_mask[0][i]
            k = asc_mask[1][i]
            scan_time = get_datetime_from_datetime64(asc_times[j][k].values)
            timestamp = scan_time.hour + scan_time.minute / 100
            timestamp_data[j][k] = timestamp

        for i in range(len(des_mask[0])):
            j = des_mask[0][i]
            k = des_mask[1][i]
            scan_time = get_datetime_from_datetime64(des_times[j][k].values)
            timestamp = scan_time.hour + scan_time.minute / 100
            timestamp_data[j][k] = timestamp

        new_xarray_dict = make_smap_dataset(
            latitude,
            longitude,
            new_smap_data,
            timestamp_data,
            datetime.strftime(split_start, utc_date_format),
        )

        new_xarray_dict.to_netcdf(awips2_file, format="NETCDF4")

        success_outputs.append(str(awips2_file))
    return success_outputs


def make_smap_data_mask(times, split_start, split_end):
    """Create a mask for only the valid times between the defined start and end.

    Parameters
    ----------
    times : xarray.Dataset
    split_start : datetime.datetime
    split_end : datetime.datetime

    Returns
    -------
    tuple[NDArray[intp], ...]
    """
    valid_mask = (
        (times >= np.datetime64(split_start))
        & (times < np.datetime64(split_end))
        & (np.logical_not(np.isnat(times)))
    )
    mask = np.where(valid_mask)
    return mask


def write_smos_awips2(xarray_dict, working_dir, utc_date_format):
    """Write AWIPS2 style NetCDF file containing SMOS data to the disk.

    Parameters
    ----------
    xarray_dict : dict[str, xarray.Dataset]
    working_dir : Path
    utc_date_format : str

    Returns
    -------
    List[str]
    """
    success_outputs = []
    dataset = xarray_dict["WINDSPEED"]
    time_array = xarray_dict["WINDSPEED"]["time"]
    wind_speed_array = xarray_dict["WINDSPEED"]["wind_speed_kts"].values
    longitude = xarray_dict["WINDSPEED"]["longitude"].values
    latitude = xarray_dict["WINDSPEED"]["latitude"].values

    quality_mask = np.where(
        (wind_speed_array >= 0.0) & (np.logical_not(np.isnan(wind_speed_array)))
    )

    new_time_data = np.full(latitude.shape, -999.0)
    new_smos_data = np.full(latitude.shape, -999.0)
    new_smos_data[quality_mask] = wind_speed_array[quality_mask]

    for i in range(len(quality_mask[0])):
        j = quality_mask[0][i]
        k = quality_mask[1][i]
        pd_time = get_datetime_from_datetime64(time_array[j][k].values)
        timestamp = pd_time.hour + (pd_time.minute / 100)
        new_time_data[j][k] = timestamp

    new_xarray_dict = make_smos_dataset(
        latitude,
        longitude,
        new_smos_data,
        new_time_data,
    )

    start_time = dataset.attrs["start_datetime"]
    # end_time = dataset.attrs["end_datetime"]
    new_xarray_dict.attrs["time_coverage_start"] = dataset.attrs["time_coverage_start"]
    new_xarray_dict.attrs["time_coverage_end"] = dataset.attrs["time_coverage_end"]
    new_xarray_dict.attrs["time_coverage_resolution"] = dataset.attrs[
        "time_coverage_resolution"
    ]
    new_xarray_dict.attrs["answrs:acquistion_time"] = start_time.strftime(
        utc_date_format
    )

    base_file = "CIRA_A2WIND_SMOS_" + start_time.strftime("%Y%m%d_%H%M")
    awips2_file = Path(working_dir, f"{base_file}.nc")

    new_xarray_dict.to_netcdf(awips2_file, format="NETCDF4")

    success_outputs.append(str(awips2_file))

    return success_outputs


def make_smos_dataset(lat, lon, smos_data, timestamp_data):
    """Create an xarray.Dataset with the final SMOS data.

    Parameters
    ----------
    lat : xarray.Dataset or xarray.DataArray
    lon : xarray.Dataset or xarray.DataArray
    smos_data : np.ndarray with the smos windspeeds
    timestamp_data : np.ndarray with valid timestamps
    acquisition_time : str datetime formatted for AWIPS2

    Returns
    -------
    new_dataset : xarray.Dataset
    """
    latitude = xr.DataArray(
        lat,
        dims=("y", "x"),
        attrs={
            "units": "degrees_north",
            "long_name": "Latitude array in decimal degrees",
            "standard_name": "latitude",
        },
    )
    longitude = xr.DataArray(
        lon,
        dims=("y", "x"),
        attrs={
            "units": "degrees_east",
            "long_name": "Longitude array in decimal degrees",
            "standard_name": "longitude",
        },
    )
    smos_wind = xr.DataArray(
        smos_data,
        dims=("y", "x"),
        attrs={
            "units": "m s-1",
            "long_name": "SMOS-derived wind speed at 10-m height neutral stability",
            "standard_name": "wind_speed",
            "scale_factor": 1.0,
            "add_offset": 0.0,
            "coverage_content_type": "physicalMeasurement",
            "_FillValue": -999.0,
            "valid_min": 0.0,
            "valid_max": 100.0,
            "grid_mapping": "NONE",
            "source": "SMOS imagery European Space Agency (ESA)",
            "reference": "NONE",
            "platform": "SMOS WEST",
            "instrument": "SMOS",
            "Comment": "NONE",
            "coordinates": "longitude latitude",
        },
    ).astype(np.float32)
    timestamps = xr.DataArray(
        timestamp_data,
        dims=("y", "x"),
        attrs={
            "name": "SMAP_TIME",
            "long_name": "time of observation at nadir",
            "coordinates": "longitude latitude",
            "units": "Z",
            "_FillValue": -999.0,
        },
    ).astype(np.float32)
    new_dataset = xr.Dataset(
        {
            "latitude": latitude,
            "longitude": longitude,
            "smos_wind": smos_wind,
            "TTIME": timestamps,
        },
        attrs={
            "title": "SMOS winds at 10-m height neutral stability",
            "source": "SMOS imagery European Space Agency (ESA)",
            "Conventions": "CF-1.6",
            "license": "These data are available for use without restriction",
            "sensor": "SMOS",
            "grid_mapping": "Ground Projection samples in meters",
            "keywords": "Ocean > Wind Speed",
            "keywords_vocabulary": "UNKNOWN",
            "standard_name_vocabulary": "UNKNOWN",
            "process_level": "Level 2",
            "processing_level": "Level 2",
            "cdm_data_type": "Grid",
            "feature_type": "Grid",
            "summary": "'SMOS wind measurements'",
            "netcdf_version_id": "4.3.2",
            "rows": " 721.",
            "columns": " 1440.",
            "geospatial_vertical_min": "10",
            "geospatial_vertical_max": "10",
            "geospatial_units": "meters",
            "geospatial_resolution": "1",
            "answrs:answrs_product_name": "SMOS-derived wind speed",
            "answrs:platform_name": "SMOS",
        },
    )

    return new_dataset


def make_smap_dataset(
    lat: xr.DataArray,
    lon: xr.DataArray,
    new_smap_data: np.ndarray,
    timestamp_data: np.ndarray,
    acquisition_time: str,
):
    """Create an xarray.Dataset with the final SMAP data.

    Parameters
    ----------
    lat : xarray.Dataset or xarray.DataArray
    lon : xarray.Dataset or xarray.DataArray
    new_smap_data : np.ndarray with the smap windspeeds
    timestamp_data : np.ndarray with valid timestamps
    acquisition_time : str datetime formatted for AWIPS2

    Returns
    -------
    new_dataset : xarray.Dataset
    """
    latitude = xr.DataArray(
        lat,
        dims=("y", "x"),
        attrs={
            "units": "degrees_north",
            "long_name": "Latitude array in decimal degrees",
            "standard_name": "latitude",
        },
    )
    longitude = xr.DataArray(
        lon,
        dims=("y", "x"),
        attrs={
            "units": "degrees_east",
            "long_name": "Longitude array in decimal degrees",
            "standard_name": "longitude",
        },
    )
    smap_wind = xr.DataArray(
        new_smap_data,
        dims=("y", "x"),
        attrs={
            "units": "kts",
            "long_name": "SMAP-derived wind speed at 10-m height neutral stability",
            "standard_name": "wind_speed",
            "scale_factor": 1.0,
            "add_offset": 0.0,
            "coverage_content_type": "physicalMeasurement",
            "_FillValue": -999.0,
            "valid_min": 0.0,
            "valid_max": 100.0,
            "grid_mapping": "NONE",
            "source": "RSS",
            "reference": "NONE",
            "platform": "SMAP",
            "instrument": "SMAP",
            "coordinates": "longitude latitude",
        },
    ).astype(np.float32)
    timestamps = xr.DataArray(
        timestamp_data,
        dims=("y", "x"),
        attrs={
            "name": "SMAP_TIME",
            "long_name": "time of observation at nadir",
            "coordinates": "longitude latitude",
            "units": "Z",
            "_FillValue": -999.0,
        },
    ).astype(np.float32)
    new_dataset = xr.Dataset(
        {
            "latitude": latitude,
            "longitude": longitude,
            "smap_wind": smap_wind,
            "TTIME": timestamps,
        },
        attrs={
            "title": "SMAP winds at 10-m height neutral stability",
            "source": "RSS/CIRA",
            "answrs:answrs_product_name": "SMAP-derived wind speed",
            "answrs:platform_name": "SMAP",
            "answrs:acquistion_time": acquisition_time,
            "answrs:acq_date_time_begin": acquisition_time,
        },
    )

    return new_dataset


def get_min_max_times_from_two_nodes(
    ascending_nodes: xr.Dataset | xr.DataArray,
    descending_nodes: xr.Dataset | xr.DataArray,
    varname: str,
):
    """Find the minimum and maximum time values from two xarray objects.

    Parameters
    ----------
    ascending_nodes : xarray.Dataset or xarray.DataArray
    descending_nodes : xarray.Dataset or xarray.DataArray
    varname : Timestamp variable name from which to extract the min/max values

    Returns
    -------
    mintime : datetime.datetime
    maxtime : datetime.datetime
    """
    asc_min = get_min_from_xarray_time(ascending_nodes, varname)
    des_min = get_min_from_xarray_time(descending_nodes, varname)
    asc_max = get_max_from_xarray_time(ascending_nodes, varname)
    des_max = get_max_from_xarray_time(descending_nodes, varname)
    return min(asc_min, des_min), max(asc_max, des_max)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Data manipulation steps for standard absolute difference algorithm.

Generalized algorithm to calculate the difference between 1 or 2 variables taken from
multiple scan times over the same sector.
"""
import logging

import numpy as np
import xarray as xr

from geoips.data_manipulations.corrections import apply_data_range
from geoips.errors import PluginError

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "xarray_to_xarray"
name = "absdiff_mst"


def call(
    xobj,
    variables,
    product_name,
    output_data_range=None,
    input_units=None,
    output_units=None,
    min_outbounds="crop",
    max_outbounds="crop",
    norm=False,
    inverse=False,
    sun_zen_correction=False,
    mask_night=False,
    max_day_zen=None,
    mask_day=False,
    min_night_zen=None,
    gamma_list=None,
    scale_factor=None,
):
    """Apply data range and requested corrections to an absdiff product.

    Data manipulation steps for applying a data range and requested corrections
    to a single channel productn absdiff product.

    Parameters
    ----------
    xobj : xarray.Dataset
        * input xarray.Dataset used for data manipulation.
    variables : list of str
        * list of strings representing the variables used to produce 'product_name'
    product_name : str
        * String representing the product produced by this algorithm, using variables
          from 'xobj'
    output_data_range : list of float, default=None
        * list of min and max value for output data product.
        * This is applied LAST after all other corrections/adjustments
        * If None, use data min and max.
    input_units : str, default=None
        * Units of input data, for applying necessary conversions
        * If None, no conversion
    output_units : str, default=None
        * Units of output data, for applying necessary conversions
        * If None, no conversion
    min_outbounds : str, default='crop'
        * Method to use when applying bounds.  Valid values are:

            * retain: keep all pixels as is
            * mask: mask all pixels that are out of range
            * crop: set all out of range values to either min_val or max_val
              as appropriate
    max_outbounds : str, default='crop'
        * Method to use when applying bounds.  Valid values are:

            * retain: keep all pixels as is
            * mask: mask all pixels that are out of range
            * crop: set all out of range values to either min_val or max_val
              as appropriate
    norm : bool, default=False
        * Boolean flag indicating whether to normalize (True) or not (False)

            * If True, returned data will be in the range from 0 to 1
            * If False, returned data will be in the range from min_val to max_val
    inverse : bool, default=False
        * Boolean flag indicating whether to inverse (True) or not (False)

            * If True, returned data will be inverted
            * If False, returned data will not be inverted
    sun_zenith_correction : bool, default=False
        * Boolean flag indicating whether to apply solar zenith correction
          (True) or not (False)

            * If True, returned data will have solar zenith correction applied
              (see data_manipulations.corrections.apply_solar_zenith_correction)
            * If False, returned data will not be modified based on solar zenith
              angle

    Notes
    -----
    Order of operations, based on the passed arguments, is:

    1. Mask night
    2. Mask day
    3. Apply solar zenith correction
    4. Apply gamma values
    5. Apply scale factor
    6. Convert units
    7. Apply data range.

    NOTE: If "norm=True" is specified, the "output_data_range" will NOT
    match the actual range of the returned data, since the normalized data
    will be returned between 0 and 1.

    If you require a different order of operations than that specified within
    "single_channel" algorithm, please create a new algorithm for your desired
    order of operations.

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of appropriately scaled channel data,
        in units "output_units".
    """
    # Take the first lon / lat values as they should be the same for all time steps
    lon = np.asarray(xobj.longitude[0])  # Take the first longitude value
    lat = np.asarray(xobj.latitude[0])  # Take the first latitude value

    # Get the input variable from timestep 1 and 2
    if len(variables) == 1:
        data1 = np.asarray(xobj[variables[0]][0])
        data2 = np.asarray(xobj[variables[0]][1])
    # Otherwise get variable 1 from timestep 1 and variable 2 from timestep 2
    elif len(variables) == 2:
        data1 = np.asarray(xobj[variables[0]][0])
        data2 = np.asarray(xobj[variables[1]][1])
    else:
        raise PluginError(
            "Error: more than 2 variables were provided to absdiff_mst algorithm plugin"
            f".\n Variables = {variables}. Cannot take the absolute difference between "
            "more than two variables from two different time steps. Please provide only"
            f" 1-2 variable names for product: {product_name}."
        )
    # Take the absolute difference between these variables, such as cth - cbh
    data = np.abs(data2 - data1)

    data = apply_data_range(
        data,
        min_val=output_data_range[0],
        max_val=output_data_range[1],
        min_outbounds=min_outbounds,
        max_outbounds=max_outbounds,
        norm=norm,
        inverse=inverse,
    )

    final_xobj = xr.Dataset(
        data_vars={product_name: (["y", "x"], data)},
        coords={
            "longitude": (["y", "x"], lon),
            "latitude": (["y", "x"], lat),
        },
        attrs=xobj.attrs,
    )

    return final_xobj



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `run` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

import pytest

from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


class TestGeoipsRun(BaseCliTest):
    """Unit Testing Class for 'geoips run' Command."""

    amsr2_config_based_args = [
        "run_procflow",
        "${GEOIPS_TESTDATA_DIR}/test_data_amsr2/data/AMSR2-MBT_v2r2_GW1_s202005180620480_e202005180759470_c202005180937100.nc",  # noqa
        "--procflow",
        "config_based",
        "--output_config",
        "${GEOIPS_PACKAGES_DIR}/geoips/tests/yaml_configs/amsr2_test_no_compare.yaml",
        "--reader_kwargs",
        '{"test_arg": "Command line config-based amsr2 test arg"}',
        "--fuse_files",
        "${GEOIPS_TESTDATA_DIR}/test_data_amsr2/bg_data/ahi_20200518_0740/*",
        "--fuse_reader",
        "ahi_hsd",
        "--fuse_reader_kwargs",
        '{"test_arg": "Command line config-based ahi test arg"}',
        "--fuse_product",
        "Infrared-Gray",
        "--fuse_resampled_read",
        "True",
    ]
    new_amsr2_config_based_args = [arg for arg in amsr2_config_based_args]
    new_amsr2_config_based_args[0] = "geoips"
    new_amsr2_config_based_args.insert(1, "run")
    new_amsr2_config_based_args.insert(2, "config_based")

    abi_static_infrared_args = [
        "run_procflow",
        "$GEOIPS_TESTDATA_DIR/test_data_noaa_aws/data/goes16/20200918/1950/*",
        "--reader_name",
        "abi_netcdf",
        "--product_name",
        "Infrared",
        "--compare_path",
        "$GEOIPS_PACKAGES_DIR/geoips/tests/outputs/abi.static.<product>.imagery_annotated",  # noqa
        "--output_formatter",
        "imagery_annotated",
        "--filename_formatter",
        "geoips_fname",
        "--resampled_read",
        "--logging_level",
        "info",
        "--sector_list",
        "goes_east",
    ]
    new_abi_static_infrared_args = [arg for arg in abi_static_infrared_args]
    new_abi_static_infrared_args[0] = "geoips"
    new_abi_static_infrared_args.insert(1, "run")
    new_abi_static_infrared_args.insert(2, "single_source")

    geo_args = [
        "data_fusion_procflow",
        "--compare_path",
        "$GEOIPS_PACKAGES_DIR/data_fusion/tests/outputs/${curr_product}_image",  # noqa
        "--filename_formatter",
        "geoips_fname",
        "--sector_list",
        "global",
        "--fusion_final_output_formatter",
        "imagery_annotated",
        "--fusion_final_product_name",
        "Blended-Infrared-Gray",
        "--fusion_final_source_name",
        "stitched",
        "--fusion_final_platform_name",
        "geo",
        "--fuse_files",
        "$GEOIPS_TESTDATA_DIR/test_data_fusion/data/goes16_20210929.0000/*",  # noqa
        "--fuse_reader_name",
        "abi_netcdf",
        "--fuse_product_name",
        "${curr_product}",
        "--fuse_dataset_name",
        "goes16",
        "--fuse_order",
        "0",
        "--fuse_files",
        "$GEOIPS_TESTDATA_DIR/test_data_fusion/data/goes17_20210929.0000/*",  # noqa
        "--fuse_reader_name",
        "abi_netcdf",
        "--fuse_product_name",
        "${curr_product}",
        "--fuse_dataset_name",
        "goes17",
        "--fuse_order",
        "1",
        "--fuse_files",
        "$GEOIPS_TESTDATA_DIR/test_data_fusion/data/himawari8_20210929.0000/*",  # noqa
        "--fuse_reader_name",
        "ahi_hsd",
        "--fuse_product_name",
        "${curr_product}",
        "--fuse_dataset_name",
        "ahi",
        "--fuse_order",
        "2",
    ]
    new_geo_args = [arg for arg in geo_args]
    new_geo_args[0] = "geoips"
    new_geo_args.insert(1, "run")
    new_geo_args.insert(2, "data_fusion")

    @property
    def command_combinations(self):
        """A stochastic list of commands used by the GeoipsRun command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            self._cmd_list = []
            base_args = ["geoips", "run"]
            # add argument lists for legacy calls 'run_procflow' and
            # 'data_fusion_procflow'
            # additionally add argument lists for 'geoips run config_based',
            # 'geoips run data_fusion', and 'geoips run single_source'

            self._cmd_list.append(self.amsr2_config_based_args)
            self._cmd_list.append(self.new_amsr2_config_based_args)
            self._cmd_list.append(self.abi_static_infrared_args)
            self._cmd_list.append(self.new_abi_static_infrared_args)
            if "data_fusion" in self.plugin_package_names:
                # Only add these argument lists if data_fusion is installed
                self._cmd_list.append(self.geo_args)
                self._cmd_list.append(self.new_geo_args)

            # Add argument list to retrieve help message
            self._cmd_list.append(["run_procflow", "-h"])
            self._cmd_list.append(base_args + ["-h"])
            self._cmd_list.append(base_args + ["config_based", "-h"])
            self._cmd_list.append(base_args + ["single_source", "-h"])
            if "data_fusion" in self.plugin_package_names:
                # Only add these argument lists if data_fusion is installed
                # Doing this as a separate conditional to keep the arg-lists in order
                self._cmd_list.append(["data_fusion_procflow", "-h"])
                self._cmd_list.append(base_args + ["data_fusion", "-h"])
            # Add argument list with non existent arguments
            self._cmd_list.append(base_args + ["not_a_procflow"])
        return self._cmd_list

    def check_error(self, args, error):
        """Ensure that the 'geoips run ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        # An error occurred using args. Assert that args is not valid and check the
        # output of the error.
        assert (
            "To use, type `geoips run" in error
            or "'geoips run' was called alongside procflow:" in error
        )

    def check_output(self, args, output):
        """Ensure that the 'geoips run ...' successful output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        # The args provided are valid, so test that the output is actually correct
        if "-h" in args and len(args) <= 4:
            assert "To use, type `geoips run" in output
        else:
            # Checking that output from geoips run command reports succeeds
            if "single_source" in args:
                assert "Starting single_source procflow..." in output
            elif "config_based" in args:
                assert "Starting config_based procflow..."
            elif "data_fusion" in args:
                assert "Starting data_fusion procflow..." in output


test_sub_cmd = TestGeoipsRun()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips run ...' commands.

    This test covers every valid combination of commands for the 'geoips run'
    command. We also test invalid commands, to ensure that the proper help documentation
    is provided for those using the command incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with (ie. ['geoips', 'run'])
    """
    test_sub_cmd.test_command_combinations(monkeypatch, args)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Data manipulation steps for "Night_Vis" product, GeoIPS 1 Version.

This algorithm expects one VIIRS channel (DNBRad) for a single channel image.
"""
import logging
from geoips.data_manipulations.corrections import mask_day
from geoips.data_manipulations.corrections import apply_data_range, apply_gamma

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "list_numpy_to_numpy"
name = "Night_Vis_GeoIPS1"


def call(arrays, min_outbounds="crop", max_outbounds="crop", max_night_zen=90):
    """Night Vis product algorithm data manipulation steps, GeoIPS 1 version.

    This algorithm expects DNBRad in reflectance, and returns the adjusted
    array.

    Parameters
    ----------
    arrays : list of numpy.ndarray
        * list of numpy.ndarray or numpy.MaskedArray of channel data,
            in order of sensor "channels" list
        * Degrees Kelvin

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of adjusted DNB output.

    Notes
    -----
    It will generate a product in daytime if we do not apply the daytime check.
    For now, it is for both day/night.

    We will decide whether this product is only for nighttime.
    If so, a daytime check will be required.

    We may focus only on nighttime product with moonlight after additional
    validation (TBD).
    """
    data = arrays[0]
    sun_zenith = arrays[1]

    data = mask_day(data, sun_zenith, max_night_zen)

    data_range1 = [5.0e-10, 2.5e-8]

    data = apply_data_range(
        data,
        min_val=data_range1[0],
        max_val=data_range1[1],
        min_outbounds=min_outbounds,
        max_outbounds=max_outbounds,
        norm=True,
        inverse=False,
    )
    data = apply_gamma(data, 2.0)

    return data



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Test script for representative product comparisons."""

import subprocess
import logging
from os.path import splitext
from shutil import copy
from os.path import exists, join
from importlib.resources import files
import os

from geoips.commandline.log_setup import log_with_emphasis
from geoips.geoips_utils import get_numpy_seeded_random_generator

predictable_random = get_numpy_seeded_random_generator()

LOG = logging.getLogger(__name__)

interface = "output_checkers"
family = "standard"
name = "text"


def write_percent_of_list_elements_to_file_and_return_filepath(
    list_with_lines,
    filename,
    percent_to_write,
    random_func=predictable_random.random,
):
    """Grabs a random subset of lines in a file and writes them."""
    threshold = float(100 - percent_to_write) / 100
    with open(filename, "w") as f:
        for line in list_with_lines:
            if random_func() > threshold:
                f.write(line)


def get_test_files(test_data_dir):
    """Return a series of varied text files."""
    save_dir = join(test_data_dir, "scratch", "unit_tests", "test_text")
    if not exists(save_dir):
        os.mkdir(save_dir, exists_ok=True)

    text_path = str(files("geoips") / "plugins/txt/ascii_palettes/tpw_cimss.txt")

    comp_path = join(save_dir, "compare.txt")
    copy(text_path, comp_path)
    with open(comp_path, mode="r") as comp_txt:
        comp_text_contents = comp_txt.readlines()

    return comp_path, [
        write_percent_of_list_elements_to_file_and_return_filepath(
            comp_text_contents, join(save_dir, "matched.txt"), 100
        ),
        write_percent_of_list_elements_to_file_and_return_filepath(
            comp_text_contents, join(save_dir, "slightly_mismatch.txt"), 95
        ),
        write_percent_of_list_elements_to_file_and_return_filepath(
            comp_text_contents, join(save_dir, "bad_mismatch.txt"), 75
        ),
    ]


def perform_test_comparisons(plugin, compare_file, test_files):
    """Test the comparison of two text files with the Text Output Checker."""
    for path_idx in range(len(test_files)):
        retval = plugin.module.outputs_match(
            plugin,
            test_files[path_idx],
            compare_file,
        )
        if path_idx == 0:
            assert retval is True
        else:
            assert retval is False


def correct_file_format(fname):
    """Check if fname is a text file.

    Parameters
    ----------
    fname : str
        Name of file to check.

    Returns
    -------
    bool
        True if it is a text file, False otherwise.
    """
    if splitext(fname)[-1] in ["", ".txt", ".text", ".yaml"]:
        with open(fname) as f:
            line = f.readline()
        if isinstance(line, str):
            return True
    return False


def outputs_match(plugin, output_product, compare_product):
    """Check if two text files match.

    Parameters
    ----------
    plugin: OutputCheckerPlugin
        The corresponding geotiff OutputCheckerPlugin that has access to needed methods
    output_product : str
        Full path to current output product
    compare_product : str
        Full path to "good" comparison product

    Returns
    -------
    bool
        Return True if products match, False if they differ
    """
    ret = subprocess.run(
        ["diff", output_product, compare_product],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    LOG.debug(ret.stdout)
    if ret.returncode == 0:
        log_with_emphasis(LOG.info, "GOOD Text files match")
        return True

    out_difftxt = plugin.get_out_diff_fname(compare_product, output_product)
    with open(out_difftxt, "w") as fobj:
        subprocess.call(["diff", output_product, compare_product], stdout=fobj)
    log_with_emphasis(
        LOG.interactive,
        "BAD Text files do NOT match exactly",
        f"output_product: {output_product}",
        f"compare_product: {compare_product}",
        f"out_difftxt: {out_difftxt}",
    )
    return False


def call(plugin, compare_path, output_products):
    """Compare the "correct" text found the list of current output_products.

    Compares files produced in the current processing run with the list of
    "correct" files contained in "compare_path".

    Parameters
    ----------
    plugin: OutputCheckerPlugin
        The corresponding text OutputCheckerPlugin that has access to needed methods
    compare_path : str
        Path to directory of "correct" products - filenames must match output_products
    output_products : list of str
        List of strings of current output products,
        to compare with products in compare_path

    Returns
    -------
    int
        Binary code: 0 if all comparisons were completed successfully.
    """
    retval = plugin.compare_outputs(
        compare_path,
        output_products,
    )
    return retval



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Utilities for merging granules into a single data array.

These utilities can apply to potentially different data sources -
spanning a variety of sensors and platforms into a single final
dataset.
"""
# Python Standard Libraries
import logging
from datetime import timedelta

# Third Party Installed Libraries

LOG = logging.getLogger(__name__)


def minrange(start_date, end_date):
    """Check one minute at a time."""
    # log.info('in minrange')
    tr = end_date - start_date
    mins = int((tr.seconds + tr.days * 86400) / 60)
    LOG.info("%s", mins)
    for n in range(mins):
        yield start_date + timedelta(seconds=(n * 60))


def daterange(start_date, end_date):
    """Check one day at a time.

    If end_date - start_date is between 1 and 2, days will be 1,
    and range(1) is 0. So add 2 to days to set range.
    """
    # log.info('in minrange')
    tr = end_date - start_date
    for n in range(tr.days + 2):
        yield start_date + timedelta(n)


def hourrange(start_date, end_date):
    """Check one hour at a time."""
    LOG.info("in hourrange")
    tr = end_date - start_date
    for n in range(tr.days * 24 + tr.seconds / 3600):
        yield start_date + timedelta(seconds=(n * 3600))


def find_datafiles_in_range(
    sector_name,
    platform_name,
    source_name,
    min_time,
    max_time,
    basedir,
    product_name,
    every_min=True,
    verbose=False,
    time_format="%H%M",
    actual_datetime=None,
    single_match=False,
):
    """Find datafiles from a specified set of parameters.

    Parameters
    ----------
    sector_name : str
        Sector of interest
    platform_name : str
        platform of interest
    source_name : str
        Source of interest
    min_time : datetime.datetime
        Minimum time to search
    max_time : datetime.datetime
        Maximum time to search
    basedir : str
        Base directory to search
    product_name : str
        Product of interest
    every_min : bool, optional
        Check every minute, by default True
    verbose : bool, optional
        Print a lot of log output during the search, by default False
    time_format : str, optional
        Format of time information in filenames, by default "%H%M"
    actual_datetime : datetime.datetime, optional
        Actual datetime of the requested data, required if single_match is
        True, by default None
    single_match : bool, optional
        Only return the closest matching file if True, else return all
        matching files, by default False

    Returns
    -------
    list of str
        List of all filenames matching the given parameters
        (list of length 1 if single_match is True, all matching files
        if single_match is false)
    """
    LOG.info(
        "Finding %s %s %s files between %s and %s",
        sector_name,
        platform_name,
        source_name,
        min_time,
        max_time,
    )
    from glob import glob
    from geoips.filenames.product_filenames import netcdf_write_filename

    fnames = []
    first = True
    min_timediff = 10000000
    if (min_time - max_time) < timedelta(minutes=30) or every_min is True:
        for sdt in minrange(min_time, max_time):
            ncdf_fname = netcdf_write_filename(
                basedir,
                product_name=product_name,
                source_name=source_name,
                platform_name=platform_name,
                sector_name=sector_name,
                product_datetime=sdt,
                time_format=time_format,
            )
            ncdf_fnames = glob(ncdf_fname)
            if first:
                LOG.info("First path: %s", ncdf_fname)
                first = False
            if verbose:
                LOG.info("Checking %s", ncdf_fname)
            if ncdf_fnames:
                if (
                    single_match is True
                    and abs((sdt - actual_datetime).seconds) < min_timediff
                ):
                    min_timediff = (sdt - actual_datetime).seconds
                    LOG.info(
                        "    Adding %s, min_timediff now %s", ncdf_fnames, min_timediff
                    )
                    fnames = ncdf_fnames
                if single_match is not True:
                    LOG.info("    Adding %s", ncdf_fnames)
                    fnames += ncdf_fnames

    return fnames


def get_matching_files(
    primary_sector_name,
    subsector_names,
    platforms,
    sources,
    max_time_diffs,
    basedir,
    merge_datetime,
    product_name,
    time_format="%H%M",
    buffer_mins=30,
    verbose=False,
    single_match=False,
):
    """Given the current set of parameters, find all matching files.

    Given the current primary sector, and associated subsectors, platforms,
    and sources, find all matching files.

    Parameters
    ----------
    primary_sector_name : str
        The final sector that all data will be stitched into.
        ie 'GlobalGlobal'
    subsector_names : list of str
        List of all subsectors that will be merged into the final sector.
        (potentially including the full primary_sector_name.)
        ie ['GlobalGlobal', 'GlobalAntarctic', 'GlobalArctic']
    platforms : list of str
        List of all desired platforms.
        platforms, sources, and max_time_diffs correspond to one another
        and should be the same length and in the same order.
    sources : list of str
        List of all desired sources.
        platforms, sources, and max_time_diffs correspond to one another
        and should be the same length and in the same order.
    max_time_diffs : list of int
        Minutes. List of allowed time diffs for given platform/source. Matches
        max_time_diff before the requested merge_datetime argument.
        platforms, sources, and max_time_diffs correspond to one another
        and should be the same length and in the same order.
    basedir : str
        Base directory in which to look for the matching files.
    merge_datetime : datetime
        Attempt matching max_time_diff prior to merge_datetime
    product_name : str
        product_name string found in matching files
    time_format : str, optional
        Requested time format for filenames (strptime format string),
        by default '%H%M'
    verbose : bool, optional
        Print a lot of log output during the search, by default False
    single_match : bool, optional
        Only return the closest matching file if True, else return all
        matching files, by default False

    Returns
    -------
    list of str
        List of all filenames matching the given parameters
        (list of length 1 if single_match is True, all matching files
        if single_match is false)
    """
    from datetime import timedelta

    # MLS1 prev_files should be a dictionary with primary sector names as keys
    # - otherwise it will not work.
    prev_files = []
    actual_datetime = None
    if single_match is True:
        actual_datetime = merge_datetime

    # Go through the list of platforms / sources / allowed time diffs to find
    # the appropriate files for each data type
    for platform, source, time_diff in zip(platforms, sources, max_time_diffs):
        # Go through the list of all subsectors needed to merge into the primary
        # sector listed above
        for currsectname in subsector_names:
            # Get the sector name from the current subsector.
            LOG.info("Checking %s %s %s %s", currsectname, platform, source, time_diff)
            curr_verbose = verbose
            # if platform == 'npp':
            #     curr_verbose=False

            # Use the time/source/platform/sector name to find the desired
            # matching files.
            # MLS1 This should be prev_files[primary_sector_name]
            prev_files += find_datafiles_in_range(
                currsectname,
                platform,
                source,
                merge_datetime - timedelta(minutes=time_diff),
                merge_datetime + timedelta(minutes=buffer_mins),
                basedir=basedir,
                product_name=product_name,
                every_min=True,
                verbose=curr_verbose,
                time_format=time_format,
                actual_datetime=actual_datetime,
                single_match=single_match,
            )
    return prev_files



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""TC trackfile parser for B-Deck formatted TC deck files.

Each B-Deck file contains the full history of storm BEST tracks, on storm per
location per line (split between 3 lines each in comments for readability)::

  AL, 20, 2020091318,   , BEST,   0, 126N,  374W,  30, 1006, TD,   0,    ,    0,
      0,    0,    0, 1011,  240, 100,  40,   0,   L,   0,    ,   0,   0,
           TWENTY, M, 12, NEQ,   60,    0,    0,   60, genesis-num, 039,
  AL, 20, 2020091400,   , BEST,   0, 130N,  386W,  30, 1006, TD,   0,    ,    0,
          0,    0,    0, 1011,  240, 100,  40,   0,   L,   0,    ,   0,   0,
           TWENTY, M, 12, NEQ,   60,   60,    0,    0, genesis-num, 039,
  AL, 20, 2020091406,   , BEST,   0, 130N,  404W,  35, 1004, TS,  34, NEQ,    0,
          0,   40,   40, 1011,  240,  40,   0,   0,   L,   0,    ,   0,   0,
            TEDDY, M,  0,    ,    0,    0,    0,    0, genesis-num, 039,
  AL, 20, 2020091412,   , BEST,   0, 128N,  422W,  35, 1004, TS,  34, NEQ,   50,
         30,    0,   50, 1011,  240,  40,  45,   0,   L,   0,    ,   0,   0,
            TEDDY, M,  0,    ,    0,    0,    0,    0, genesis-num, 039,
  AL, 20, 2020091418,   , BEST,   0, 129N,  434W,  40, 1003, TS,  34, NEQ,   80,
         40,    0,   70, 1012,  210,  50,  55,   0,   L,   0,    ,   0,   0,
            TEDDY, M, 12, NEQ,   90,   30,    0,   30, genesis-num, 039,
  AL, 20, 2020091500,   , BEST,   0, 130N,  445W,  45, 1002, TS,  34, NEQ,  100,
         50,    0,   80, 1012,  210,  40,  55,   0,   L,   0,    ,   0,   0,
            TEDDY, M, 12, NEQ,   90,   30,    0,   30, genesis-num, 039,
  AL, 20, 2020091506,   , BEST,   0, 134N,  455W,  50, 1001, TS,  34, NEQ,  100,
         50,   20,   80, 1012,  210,  20,  60,   0,   L,   0,    ,   0,   0,
            TEDDY, M, 12, NEQ,  300,  210,   30,    0, genesis-num, 039,
  AL, 20, 2020091506,   , BEST,   0, 134N,  455W,  50, 1001, TS,  50, NEQ,   20,
          0,    0,    0, 1012,  210,  20,  60,   0,   L,   0,    ,   0,   0,
            TEDDY, M, 12, NEQ,  300,  210,   30,    0, genesis-num, 039,
  AL, 20, 2020091512,   , BEST,   0, 138N,  466W,  55,  999, TS,  34, NEQ,  140,
     60,   40,  160, 1011,  250,  20,  65,   0,   L,   0,    ,   0,   0,
        TEDDY, D, 12, NEQ,  300,  210,   30,   30, genesis-num, 039,
  AL, 20, 2020091512,   , BEST,   0, 138N,  466W,  55,  999, TS,  50, NEQ,   30,
      0,    0,   30, 1011,  250,  20,  65,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  300,  210,   30,   30, genesis-num, 039,
  AL, 20, 2020091518,   , BEST,   0, 142N,  475W,  55,  997, TS,  34, NEQ,  140,
     80,   40,  160, 1011,  250,  20,  65,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  300,  240,   60,   60, genesis-num, 039,
  AL, 20, 2020091518,   , BEST,   0, 142N,  475W,  55,  997, TS,  50, NEQ,   30,
      0,    0,   30, 1011,  250,  20,  65,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  300,  240,   60,   60, genesis-num, 039,
  AL, 20, 2020091600,   , BEST,   0, 147N,  480W,  65,  987, HU,  34, NEQ,  140,
     80,   40,  150, 1010,  180,  20,  75,   0,   L,   0,    ,   0,   0,
            TEDDY, D,  0,    ,    0,    0,    0,    0, genesis-num, 039,
  AL, 20, 2020091600,   , BEST,   0, 147N,  480W,  65,  987, HU,  50, NEQ,   40,
     30,    0,   30, 1010,  180,  20,  75,   0,   L,   0,    ,   0,   0,
            TEDDY, D,  0,    ,    0,    0,    0,    0, genesis-num, 039,
  AL, 20, 2020091600,   , BEST,   0, 147N,  480W,  65,  987, HU,  64, NEQ,   20,
      0,    0,    0, 1010,  180,  20,  75,   0,   L,   0,    ,   0,   0,
            TEDDY, D,  0,    ,    0,    0,    0,    0, genesis-num, 039,
  AL, 20, 2020091606,   , BEST,   0, 154N,  486W,  80,  978, HU,  34, NEQ,  140,
     80,   40,  150, 1010,  180,  20, 100,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  300,  240,  120,  150, genesis-num, 039,
  AL, 20, 2020091606,   , BEST,   0, 154N,  486W,  80,  978, HU,  50, NEQ,   40,
     30,   20,   30, 1010,  180,  20, 100,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  300,  240,  120,  150, genesis-num, 039,
  AL, 20, 2020091606,   , BEST,   0, 154N,  486W,  80,  978, HU,  64, NEQ,   20,
     10,   10,   20, 1010,  180,  20, 100,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  300,  240,  120,  150, genesis-num, 039,
  AL, 20, 2020091612,   , BEST,   0, 161N,  493W,  85,  973, HU,  34, NEQ,  170,
    170,   40,  150, 1010,  180,  20, 100,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  270,  270,  240,  270, genesis-num, 039,
  AL, 20, 2020091612,   , BEST,   0, 161N,  493W,  85,  973, HU,  50, NEQ,   50,
     50,   20,   30, 1010,  180,  20, 100,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  270,  270,  240,  270, genesis-num, 039,
  AL, 20, 2020091612,   , BEST,   0, 161N,  493W,  85,  973, HU,  64, NEQ,   25,
     25,   10,   20, 1010,  180,  20, 100,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  270,  270,  240,  270, genesis-num, 039,
  AL, 20, 2020091618,   , BEST,   0, 168N,  502W,  85,  973, HU,  34, NEQ,  190,
    100,   70,  170, 1010,  180,  20, 105,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  300,  300,  240,  300, genesis-num, 039,
  AL, 20, 2020091618,   , BEST,   0, 168N,  502W,  85,  973, HU,  50, NEQ,   80,
     50,   30,   90, 1010,  180,  20, 105,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  300,  300,  240,  300, genesis-num, 039,
  AL, 20, 2020091618,   , BEST,   0, 168N,  502W,  85,  973, HU,  64, NEQ,   30,
     25,    0,   30, 1010,  180,  20, 105,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  300,  300,  240,  300, genesis-num, 039,
  AL, 20, 2020091700,   , BEST,   0, 174N,  511W,  85,  973, HU,  34, NEQ,  220,
    100,   80,  170, 1009,  210,  20, 100,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  330,  300,  270,  300, genesis-num, 039,
  AL, 20, 2020091700,   , BEST,   0, 174N,  511W,  85,  973, HU,  50, NEQ,   60,
     50,   50,   70, 1009,  210,  20, 100,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  330,  300,  270,  300, genesis-num, 039,
  AL, 20, 2020091700,   , BEST,   0, 174N,  511W,  85,  973, HU,  64, NEQ,   30,
     25,   20,   30, 1009,  210,  20, 100,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  330,  300,  270,  300, genesis-num, 039,
  AL, 20, 2020091706,   , BEST,   0, 180N,  520W,  85,  973, HU,  34, NEQ,  220,
    100,   80,  170, 1009,  210,  20, 105,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  330,  360,  300,  300, genesis-num, 039,
  AL, 20, 2020091706,   , BEST,   0, 180N,  520W,  85,  973, HU,  50, NEQ,   60,
     50,   50,   70, 1009,  210,  20, 105,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  330,  360,  300,  300, genesis-num, 039,
  AL, 20, 2020091706,   , BEST,   0, 180N,  520W,  85,  973, HU,  64, NEQ,   30,
     25,   20,   30, 1009,  210,  20, 105,   0,   L,   0,    ,   0,   0,
            TEDDY, D, 12, NEQ,  330,  360,  300,  300, genesis-num, 039,
"""
import os
import logging
from datetime import datetime

LOG = logging.getLogger(__name__)

interface = "sector_metadata_generators"
family = "tc"
name = "bdeck_parser"


def call(trackfile_name):
    """TC deckfile parser for B-Deck files.

    Each B-Deck file contains the full history of storm BEST tracks, one storm
    location per line. Example b-deck files are available in the GeoIPS repo.

    Parameters
    ----------
    trackfile_name : str
        Path to bdeck file, with full 6 hourly storm track
        history, formatted as follows:

    Returns
    -------
    list
        List of Dictionaries of storm metadata fields from each storm location

    See Also
    --------
    :ref:`api`
        Valid fields can be found in geoips.sector_utils.utils.SECTOR_INFO_ATTRS
    """
    LOG.info("STARTING getting fields from %s", trackfile_name)
    # Must get tcyear out of the filename in case a storm
    # crosses TC vs calendar years.
    # tcyear = os.path.basename(trackfile_name)[5:9]
    tc_year = get_stormyear_from_bdeck_filename(trackfile_name)
    # print tcyear

    flatsf_lines = open(trackfile_name).readlines()
    final_storm_name = get_final_storm_name_bdeck(flatsf_lines, tc_year, trackfile_name)
    invest_number = get_invest_number_bdeck(flatsf_lines)

    # This just pulls the time of the first entry in the deck file
    entry_storm_start_datetime = get_storm_start_datetime_from_bdeck_entry(flatsf_lines)

    # Note storms are often started with 3 locations, and the first 2 locations
    # are removed in later deck files, so initial storm start time often does
    # not match the final storm start time.
    # Keep track of the start datetime referenced in the filename
    filename_storm_start_datetime = get_storm_start_datetime_from_bdeck_filename(
        trackfile_name
    )

    LOG.info("  USING final_storm_name from bdeck %s", final_storm_name)
    LOG.info("  USING storm_start_datetime from bdeck %s", entry_storm_start_datetime)
    LOG.info(
        "  USING original_storm_start_datetime from bdeck %s",
        filename_storm_start_datetime,
    )

    # flatsf_lines go from OLDEST to NEWEST (so firsttime is the OLDEST
    # storm location)
    all_fields = []

    for line in flatsf_lines:
        curr_fields = parse_bdeck_line(
            line,
            source_filename=trackfile_name,
            storm_year=tc_year,
            final_storm_name=final_storm_name,
            invest_number=invest_number,
            storm_start_datetime=entry_storm_start_datetime,
            original_storm_start_datetime=filename_storm_start_datetime,
            parser_name="bdeck_parser",
        )
        # Was previously RE-SETTING finalstormname here.
        # That is why we were getting incorrect final_storm_name fields
        all_fields += [curr_fields]

    LOG.info("FINISHED getting fields from %s", trackfile_name)

    return all_fields, final_storm_name, tc_year


def lat_to_dec(lat_str):
    """Return decimal latitude based on N/S specified string."""
    latnodec = lat_str
    latdec = latnodec[:-2] + "." + latnodec[-2:]
    latdecsign = latdec[:-1] if (latdec[-1] == "N") else "-" + latdec[:-1]
    return latdecsign


def lon_to_dec(lon_str):
    """Return decimal longitude based on E/W specified string."""
    lonnodec = lon_str
    londec = lonnodec[:-2] + "." + lonnodec[-2:]
    londecsign = londec[:-1] if (londec[-1] == "E") else "-" + londec[:-1]
    return londecsign


def parse_bdeck_line(
    line,
    source_filename=None,
    storm_year=None,
    final_storm_name=None,
    invest_number=None,
    storm_start_datetime=None,
    original_storm_start_datetime=None,
    parser_name="bdeck_parser",
):
    """Retrieve the storm information from the current line from the deck file.

    Parameters
    ----------
    line : str
        Current line from the deck file including all storm information

        * AL, 20, 2020091618,   , BEST,   0, 168N,  502W,  85,  973, HU,  64,
          NEQ,   30,   25,    0,   30, 1010,  180,  20, 105,   0,   L,   0,    ,
          0,   0,      TEDDY, D, 12, NEQ,  300,  300,  240,  300,
          genesis-num, 039,
        * AL, 20, 2020091700,   , BEST,   0, 174N,  511W,  85,  973, HU,  34,
          NEQ,  220,  100,   80,  170, 1009,  210,  20, 100,   0,   L,   0,    ,
          0,   0,      TEDDY, D, 12, NEQ,  330,  300,  270,  300,
          genesis-num, 039,
        * AL, 20, 2020091700,   , BEST,   0, 174N,  511W,  85,  973, HU,  50,
          NEQ,   60,   50,   50,   70, 1009,  210,  20, 100,   0,   L,   0,    ,
          0,   0,      TEDDY, D, 12, NEQ,  330,  300,  270,  300,
          genesis-num, 039,

    Returns
    -------
    dict
        Dictionary of the fields from the current storm location from the
        deck file

    See Also
    --------
    :ref:`api`
        Valid fields can be found in geoips.sector_utils.utils.SECTOR_INFO_ATTRS
    """
    # This works with G (GeoIPS) Deck files.
    # Need separate parser for B decks (best tracks)
    # parts = line.split(',', 40)
    parts = [part.strip() for part in line.split(",")]
    if len(parts) != 38 and len(parts) != 30 and len(parts) != 40 and len(parts) != 42:
        LOG.interactive(source_filename)
        LOG.interactive(line)
        raise ValueError(
            "Incorrectly formatted deck file - "
            "must have either 30 or 38 or 42 fields, "
            f"had {len(parts)}",
        )
    fields = {}
    fields["deck_line"] = line.strip()
    fields["storm_basin"] = parts[0]
    fields["storm_num"] = int(parts[1])
    fields["synoptic_time"] = datetime.strptime(parts[2], "%Y%m%d%H")

    if isinstance(storm_start_datetime, datetime):
        fields["storm_start_datetime"] = storm_start_datetime
    if isinstance(original_storm_start_datetime, datetime):
        fields["original_storm_start_datetime"] = original_storm_start_datetime

    fields["aid_type"] = parts[
        4
    ]  # BEST, MBAM, OFCL, JTWC, etc - BEST202101220600 when updated
    # CARQ - not best track, real time, A-deck (Aids), F (Fix), E (Error), B (Best)
    fields["clat"] = float(lat_to_dec(parts[6]))
    fields["clon"] = float(lon_to_dec(parts[7]))
    fields["vmax"] = parts[8]
    if fields["vmax"]:
        fields["vmax"] = float(fields["vmax"])
    fields["pressure"] = parts[9]
    if fields["pressure"]:
        fields["pressure"] = float(fields["pressure"])

    fields["storm_name"] = parts[27]
    fields["final_storm_name"] = "unknown"
    if final_storm_name:
        LOG.debug("USING passed final_storm_name %s", final_storm_name)
        fields["final_storm_name"] = final_storm_name
    else:
        LOG.debug("USING storm_name as final_storm_name %s", fields["storm_name"])
        fields["final_storm_name"] = fields["storm_name"]
    if storm_year:
        fields["storm_year"] = storm_year
    if parts[-3] == "TRANSITIONED":
        # shB02021
        invest_id = parts[-2].split(" ")[0]
        invest_year = invest_id[4:]
        invest_basin = invest_id[0:2]
        invest_num = "9" + invest_id[3]
        # tc2021sh90invest
        fields["invest_storm_id"] = (
            "tc" + invest_year + invest_basin + invest_num + "invest"
        )
    fields["invest_number"] = None

    if invest_number:
        fields["invest_number"] = invest_number

    if source_filename:
        from geoips.geoips_utils import replace_geoips_paths

        if isinstance(source_filename, str):
            fields["source_filename"] = replace_geoips_paths(source_filename)
        else:
            fields["source_filename"] = source_filename
    fields["parser_name"] = parser_name
    return fields


def get_invest_number_bdeck(deck_lines):
    """Get invest number from full bdeck file."""
    invest_number = None
    for line in deck_lines:
        fields = parse_bdeck_line(line)
        if fields["storm_name"] == "INVEST" and fields["storm_num"] > 89:
            invest_number = fields["storm_num"]
        if "invest_storm_id" in fields:
            invest_number = int(fields["invest_storm_id"][8:10])
    return invest_number


def get_storm_start_datetime_from_bdeck_entry(deck_lines):
    """Get storm start datetime from full bdeck file."""
    # Return the synoptic time of the first bdeck entry
    fields = parse_bdeck_line(deck_lines[0])
    LOG.info("  GETTING storm start time from bdeck entry %s", fields["synoptic_time"])
    return fields["synoptic_time"]


def get_storm_start_datetime_from_bdeck_filename(bdeck_filename):
    """Get storm start datetime from bdeck file name."""
    # Return the synoptic time found in the actual filename, if it exists!
    # This will ONLY be the case for INVESTS, which can use the start
    # Gwp912022.2022101400.dat
    bdeck_parts = os.path.basename(bdeck_filename).split(".")
    storm_start_datetime = None
    if len(bdeck_parts) > 2:
        try:
            storm_start_datetime = datetime.strptime(bdeck_parts[1], "%Y%m%d%H")
            LOG.info(
                "  USING storm start time found in filename %s", storm_start_datetime
            )
        except ValueError:
            LOG.warning(
                "  SKIPPING no valid storm start time found in filename, %s",
                "using first entry in bdeck",
            )
            storm_start_datetime = None
    return storm_start_datetime


def get_stormyear_from_bdeck_filename(bdeck_filename):
    """Get the storm year from the B-deck filename.

    Parameters
    ----------
    bdeck_filename : str
        * Path to deck file to search for storm year
        * Must be of format: xxxxxYYYY.dat - pulls YYYY from filename based on
          location

    Returns
    -------
    int
        Storm year
    """
    return int(os.path.basename(bdeck_filename)[5:9])


def get_final_storm_name_bdeck(deck_lines, tcyear, trackfile_name=None):
    """Get final storm name from full bdeck file."""
    final_storm_name = "INVEST"
    for line in deck_lines:
        # curr_fields = parse_bdeck_line(line, tcyear, finalstormname)
        curr_fields = parse_bdeck_line(
            line,
            storm_year=tcyear,
            final_storm_name=final_storm_name,
            source_filename=trackfile_name,
        )
        # if curr_fields['storm_name']:
        if curr_fields["storm_name"] and curr_fields["storm_name"] != "INVEST":
            LOG.debug("UPDATING final_storm_name to %s", curr_fields["storm_name"])
            final_storm_name = curr_fields["storm_name"]
    return final_storm_name



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit Tests for log output for optional dependencies found throughout geoips."""

import logging
import pytest
from geoips.utils import context_managers

opt_importer = context_managers.import_optional_dependencies
importer_path = context_managers.__file__

LOG = logging.getLogger(__name__)

# err_str = f"Failed to import {err.name} at {filename}:{lineno}. "
# err_str += "If you need it, install it."


@pytest.mark.parametrize("loglevel", ["info", "debug", "warning", "error"])
def test_import_failure_using_log(loglevel, caplog):
    """Ensure that the test fails and that the log output is correct."""
    caplog.set_level(getattr(logging, loglevel.upper()))
    with opt_importer(loglevel=loglevel):
        """Attempt to import a package and print to LOG.info if the import fails."""
        import non_existent_package  # noqa

    err_str = f"Failed to import non_existent_package at {importer_path}:24. "
    err_str += "If you need it, install it."

    assert err_str in caplog.text


@pytest.mark.parametrize("loglevel", ["info", "debug", "warning", "error"])
def test_import_success_using_log(loglevel, caplog):
    """Ensure that the test succeeds and that the log output is empty."""
    caplog.set_level(getattr(logging, loglevel.upper()))
    with opt_importer(loglevel=loglevel):
        """Attempt to import a package and print to LOG.info if the import fails."""
        import numpy  # noqa

    assert caplog.text == ""



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Data manipulation steps for "Natural Color" EUMETSAT RGB product.

This algorithm expects three visible reflectances for an RGB image:
* 0.6 um
* 0.8 um
* 1.6 um
"""
import logging

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "xarray_to_numpy"
name = "Natural_Color"


def call(xobj):
    """Natural Color RGB product algorithm data manipulation steps.

    This algorithm expects reflectance values for

    * Blue: 0.6 um, SEVIRI B03 Reflectances
    * Green: 0.8 um, SEVIRI B02 Reflectances
    * Red: 1.6 um, SEVIRI B01 Reflectances

    Parameters
    ----------
    arrays : list of numpy.ndarray
        * list of numpy.ndarray or numpy.MaskedArray of channel data,
            in order of sensor "channels" list
        * Degrees Kelvin

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of qualitative RGBA image output
    """
    bt01 = xobj["B01Ref"].to_masked_array()  # 0.6 um reflectances, B01 SEVIRI
    bt02 = xobj["B02Ref"].to_masked_array()  # 0.8 um reflectances, B02 SEVIRI
    bt03 = xobj["B03Ref"].to_masked_array()  # 1.6 um, B03 SEVIRI

    data_range = [0.0, 1.0]

    from geoips.data_manipulations.corrections import apply_data_range

    red = apply_data_range(
        bt03,
        min_val=data_range[0],
        max_val=data_range[1],
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )  # need inverse option?

    grn = apply_data_range(
        bt02,
        min_val=data_range[0],
        max_val=data_range[1],
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )

    blu = apply_data_range(
        bt01,
        min_val=data_range[0],
        max_val=data_range[1],
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )  # create image of deep clouds in blueish color
    # norm=True, inverse=False)    #create image of deep clouds in yellowish color

    from geoips.image_utils.mpl_utils import alpha_from_masked_arrays, rgba_from_arrays

    alp = alpha_from_masked_arrays([red, grn, blu])
    rgba = rgba_from_arrays(red, grn, blu, alp)

    return rgba



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read derived surface winds from REMSS SMAP netcdf data."""

# Python Standard Libraries
from glob import glob
import logging
from os.path import basename

# Third-Party Libraries
import xarray

# GeoIPS imports
from geoips.plugins.modules.readers.utils.remss_reader import (
    read_remss_data,
)

from geoips.xarray_utils.time import (
    get_min_from_xarray_time,
    get_max_from_xarray_time,
)

LOG = logging.getLogger(__name__)

MS_TO_KTS = 1.94384
DEG_TO_KM = 111.321

interface = "readers"
family = "standard"
name = "smap_remss_winds_netcdf"
source_names = ["smap-spd"]


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read one of SMAP derived winds from netcdf data.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    # Only SAR reads multiple files
    fname = fnames[0]
    wind_xarray = xarray.open_dataset(str(fname))
    wind_xarray.attrs["source_name"] = "smap-spd"
    wind_xarray.attrs["platform_name"] = "smap"
    wind_xarray.attrs["data_provider"] = "rss"
    wind_xarray.attrs["source_file_names"] = [basename(fname)]
    wind_xarray.attrs["interpolation_radius_of_influence"] = 20000
    wind_xarray.attrs["sample_distance_km"] = DEG_TO_KM / 4
    wind_xarray.attrs["minimum_coverage"] = 20

    LOG.info("Read data from %s", fname)

    if (
        hasattr(wind_xarray, "institution")
        and "Remote Sensing Systems" in wind_xarray.institution
    ):
        if hasattr(wind_xarray, "title") and "SMAP" in wind_xarray.title:
            wind_xarrays = read_remss_data(wind_xarray, "smap")

    for wind_xarray in wind_xarrays.values():
        LOG.info("Setting standard metadata")
        wind_xarray.attrs["start_datetime"] = get_min_from_xarray_time(
            wind_xarray, "time"
        )
        wind_xarray.attrs["end_datetime"] = get_max_from_xarray_time(
            wind_xarray, "time"
        )

        if "wind_speed_kts" in wind_xarray.variables:
            # These text files store wind speeds natively in kts
            wind_xarray["wind_speed_kts"].attrs["units"] = "kts"

        LOG.info(
            "Read data %s start_dt %s source %s platform %s data_provider %s roi "
            "%s native resolution",
            wind_xarray.attrs["start_datetime"],
            wind_xarray.attrs["source_name"],
            wind_xarray.attrs["platform_name"],
            wind_xarray.attrs["data_provider"],
            wind_xarray.attrs["interpolation_radius_of_influence"],
            wind_xarray.attrs["sample_distance_km"],
        )

    wind_xarrays["METADATA"] = wind_xarray[[]]

    return wind_xarrays


def get_test_files(test_data_dir):
    """Generate test xarray from test files for unit testing."""
    filepath = test_data_dir + "/test_data_smap/data/*.nc"
    filelist = glob(filepath)
    tmp_xr = call(filelist)
    if len(filelist) == 0:
        raise NameError("No files found")
    return tmp_xr


def get_test_parameters():
    """Generate test data key for unit testing."""
    return [{"data_key": "WINDSPEED_1", "data_var": "wind_speed_kts"}]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Command line script for kicking off geoips based procflows.

MUST call with --procflow.
"""

from datetime import datetime
from geoips.commandline.log_setup import setup_logging
from geoips.commandline.args import get_command_line_args
from geoips.interfaces import procflows


def main(get_command_line_args_func=None, ARGS=None):
    """Script to kick off processing based on command line args.

    Parameters
    ----------
    get_command_line_args_func : function, optional
        - Function to use in place of "get_command_line_args", default None
          If None, use geoips.commandline.args.get_command_line_args
    ARGS: argparse Namespace
        - The arguments provided to 'geoips run' / 'run_procflow'. If ARGS are provided,
          this means we have called this function via 'geoips run' and the arguments
          were obtained via the GeoIPS CLI alongside geoips.commandline.args:add_args
          For more info, see geoips.commandline.geoips_run.GeoipsRun:add_arguments
    """
    DATETIMES = {}
    DATETIMES["start"] = datetime.utcnow()

    if get_command_line_args_func is None:
        get_command_line_args_func = get_command_line_args
    # arglist=None allows all possible arguments.
    if ARGS is None:
        ARGS = get_command_line_args_func(
            arglist=None,
            description="Run data file processing",
        )

    COMMAND_LINE_ARGS = ARGS.__dict__
    if COMMAND_LINE_ARGS.get("logging_level"):
        LOG = setup_logging(logging_level=COMMAND_LINE_ARGS["logging_level"])
    else:
        LOG = setup_logging()
    LOG.info("RETRIEVED COMMAND LINE ARGUMENTS")
    LOG.interactive("\n\n\nStarting %s procflow...\n\n", COMMAND_LINE_ARGS["procflow"])
    import sys

    LOG.info(
        "COMMANDLINE CALL: \n    %s",
        "\n        ".join([currarg + " \\" for currarg in sys.argv]),
    )

    # LOG.info(COMMAND_LINE_ARGS)
    LOG.info("GETTING PROCFLOW MODULE")
    PROCFLOW = procflows.get_plugin(COMMAND_LINE_ARGS["procflow"])

    LOG.info(f"CALLING PROCFLOW MODULE: {PROCFLOW.name}")
    if PROCFLOW:
        LOG.info(COMMAND_LINE_ARGS["filenames"])
        LOG.interactive(
            "\n\n\nRunning on filenames: %s\n\n", COMMAND_LINE_ARGS["filenames"]
        )
        LOG.info(COMMAND_LINE_ARGS)
        LOG.info(PROCFLOW)
        RETVAL = PROCFLOW(COMMAND_LINE_ARGS["filenames"], COMMAND_LINE_ARGS)
        LOG.interactive(
            "Completed geoips PROCFLOW %s processing, done!",
            COMMAND_LINE_ARGS["procflow"],
        )
        LOG.info("Starting time: %s", DATETIMES["start"])
        LOG.info("Ending time: %s", datetime.utcnow())
        LOG.interactive("Total time: %s", datetime.utcnow() - DATETIMES["start"])
        if isinstance(RETVAL, list):
            for ret in RETVAL:
                LOG.interactive("GEOIPSPROCFLOWSUCCESS %s", ret)
            if len(RETVAL) > 2:
                LOG.interactive(
                    "GEOIPSTOTALSUCCESS %s %s products generated, total time %s",
                    str(PROCFLOW.name),
                    len(RETVAL),
                    datetime.utcnow() - DATETIMES["start"],
                )
            else:
                LOG.interactive(
                    "GEOIPSNOSUCCESS %s %s products generated, total time %s",
                    str(PROCFLOW.name),
                    len(RETVAL),
                    datetime.utcnow() - DATETIMES["start"],
                )
            sys.exit(0)
        # LOG.info('Return value: %s', bin(RETVAL))
        LOG.interactive("Return value: %d", RETVAL)
        sys.exit(RETVAL)

    else:
        raise IOError(
            "FAILED no geoips*/{0}.py with def {0}".format(
                COMMAND_LINE_ARGS["procflow"]
            )
        )


if __name__ == "__main__":
    main()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Title formatters interface module."""

from geoips.interfaces.base import BaseModuleInterface


class TitleFormattersInterface(BaseModuleInterface):
    """Interface for creating GeoIPS formatted titles."""

    name = "title_formatters"
    required_args = {"standard": []}
    required_kwargs = {"standard": []}


title_formatters = TitleFormattersInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Matplotlib information for standard imagery with an existing system colormap."""
import logging

from matplotlib.colors import Normalize
from matplotlib import pyplot as plt

from geoips.image_utils.colormap_utils import from_ascii
from geoips.geoips_utils import find_ascii_palette
from geoips.interfaces import colormappers

LOG = logging.getLogger(__name__)

interface = "colormappers"
family = "matplotlib"
name = "matplotlib_linear_norm"


def call(
    data_range=None,
    cmap_name="Greys",
    cmap_source="matplotlib",
    cmap_path=None,
    create_colorbar=True,
    cbar_label=None,
    cbar_ticks=None,
    cbar_tick_labels=None,
    cbar_spacing="proportional",
    cbar_full_width=False,
    colorbar_kwargs=None,
    set_ticks_kwargs=None,
    set_label_kwargs=None,
):
    """Set the matplotlib colors information for matplotlib linear norm cmaps.

    This information used in both colorbar and image production throughout
    GeoIPS image output specifications.

    Parameters
    ----------
    data_range : list, default=None
        * [min_val, max_val], matplotlib.colors.Normalize(vmin=min_val, vmax=max_val)
        * If data_range not specified, vmin and vmax both None.
    cmap_name : str, default="Greys"
        * Specify the name of the resulting matplotlib colormap.
        * If no ascii_path specified, will use builtin matplotlib
          colormap of name cmap_name.
    ascii_path : str, default=None
        * Specify full path to ASCII palette to convert to matplotlib
          colormap.
        * If not specified, use internal matplotlib colormap "cmap_name".
    cbar_label : str, default=None
        * Positional parameter passed to cbar.set_label
        * If specified, use cbar_label string as colorbar label.
    create_colorbar : bool, default=True
        * Specify whether the image should contain a colorbar or not.
    cbar_ticks : list, default=None
        * Positional parameter passed to cbar.set_ticks
        * Specify explicit list of ticks to include for colorbar.
        * None indicates ticks at int(min) and int(max) values
    cbar_tick_labels : list, default=None
        * "labels" argument to pass to cbar.set_ticks.
        * can also specify directly within "set_ticks_kwargs"
    cbar_spacing : string, default="proportional"
        * "spacing" argument to pass to fig.colorbar
        * can also specify directly within "colorbar_kwargs"
    cbar_full_width : bool, default=True
        * Extend the colorbar across the full width of the image.
    colorbar_kwargs : dict, default=None
        * keyword arguments to pass through directly to "fig.colorbar"
    set_ticks_kwargs : dict, default=None
        * keyword arguments to pass through directly to "cbar.set_ticks"
    set_label_kwargs : dict, default=None
        * keyword arguments to pass through directly to "cbar.set_label"

    Returns
    -------
    mpl_colors_info : dict
        * Specifies matplotlib Colors parameters for use in both plotting and
          colorbar generation

    See Also
    --------
    :ref:`api`
        See geoips.image_utils.mpl_utils.create_colorbar for field descriptions.
    """
    min_val = None
    max_val = None
    if data_range is not None:
        min_val = data_range[0]
        max_val = data_range[1]

    if cmap_source == "matplotlib":
        try:
            mpl_cmap = plt.get_cmap(cmap_name)
        except ValueError:
            raise ValueError(f"Colormap {cmap_name} not found in source {cmap_source}")
    elif cmap_source == "geoips":
        cmap_plugin = colormappers.get_plugin(cmap_name)
        # Just get the cmap out of mpl_colors_info to use here.
        mpl_cmap = cmap_plugin()["cmap"]
    elif cmap_source == "ascii":
        if cmap_path is not None:
            mpl_cmap = from_ascii(cmap_path, cmap_name=cmap_name)
        else:
            try:
                ascii_path = find_ascii_palette(cmap_name)
                mpl_cmap = from_ascii(ascii_path, cmap_name=cmap_name)
            except ValueError:
                raise ValueError(
                    "Colormap {cmap_name} not found in source {cmap_source}"
                )
    else:
        raise ValueError(
            "Uknown colormap source {cmap_source}, must be one of "
            "'matplotlib', 'geoips', or 'ascii'"
        )

    LOG.info("Setting norm")

    mpl_norm = Normalize(vmin=min_val, vmax=max_val)
    if cbar_ticks:
        mpl_ticks = cbar_ticks
    elif min_val is not None:
        mpl_ticks = [int(min_val), int(max_val)]
    else:
        mpl_ticks = None

    if cbar_tick_labels is None:
        mpl_tick_labels = mpl_ticks

    # Must be uniform or proportional, None not valid for Python 3
    mpl_boundaries = None

    mpl_colors_info = {
        "cmap": mpl_cmap,
        "norm": mpl_norm,
        "boundaries": mpl_boundaries,
        "colorbar": create_colorbar,
        "cbar_ticks": mpl_ticks,
        "cbar_tick_labels": mpl_tick_labels,
        "cbar_label": cbar_label,
        "cbar_spacing": cbar_spacing,
        "cbar_full_width": cbar_full_width,
        "colorbar_kwargs": colorbar_kwargs,
        "set_ticks_kwargs": set_ticks_kwargs,
        "set_label_kwargs": set_label_kwargs,
    }

    return mpl_colors_info
